-----------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Test bench for mms_diag_block_gen
--   The tb is self-stopping and self-checking.
-- Usage:
--   > as 8
--   > run -all
--   Observe tb_state and check the out_sosi_0 fields if data is OK.

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;   
USE IEEE.numeric_std.ALL;
USE std.textio.all;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE common_lib.common_lfsr_sequences_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE dp_lib.tb_dp_pkg.ALL;
USE work.diag_pkg.ALL;

ENTITY tb_mms_diag_block_gen IS
  GENERIC (
    g_use_usr_input        : BOOLEAN := TRUE;
    g_use_bg               : BOOLEAN := FALSE;
    g_use_tx_seq           : BOOLEAN := TRUE;
    g_use_bg_buffer_ram    : BOOLEAN := TRUE;
    g_usr_bypass_xonoff    : BOOLEAN := FALSE;
    g_flow_control_verify  : t_dp_flow_control_enum := e_random;  -- always active, or random flow control
    g_nof_repeat           : NATURAL := 5;
    g_nof_streams          : NATURAL := 2;
    g_gap_size             : NATURAL := 0
  );
END tb_mms_diag_block_gen;

ARCHITECTURE tb OF tb_mms_diag_block_gen IS
  
  CONSTANT clk_period           : TIME := 10 ns;
  
  CONSTANT c_verify_usr_rflow   : BOOLEAN := g_flow_control_verify/=e_active;
  CONSTANT c_use_user_fifo      : BOOLEAN := c_verify_usr_rflow;
  
  CONSTANT c_file_name_prefix   : STRING  := "data/bf_in_data";
  CONSTANT c_buf_dat_w          : NATURAL := 32; 
  CONSTANT c_buf_addr_w         : NATURAL := 7;
  CONSTANT c_block_size         : NATURAL := 96;
  
  CONSTANT c_bsn_offset         : NATURAL := 42;
  CONSTANT c_sync_period        : NATURAL := 10;
  CONSTANT c_sync_offset        : NATURAL := c_bsn_offset MOD c_sync_period;
  
  CONSTANT c_period_size        : NATURAL := c_block_size + g_gap_size;
  
  CONSTANT c_mem_lo_addr        : NATURAL := 0;
  CONSTANT c_mem_hi_addr        : NATURAL := c_block_size-1;
  
  CONSTANT c_pend_period        : NATURAL :=  1 * c_period_size;
  CONSTANT c_on_period          : NATURAL := 10 * c_period_size;
  CONSTANT c_off_period         : NATURAL := 10 * c_period_size;
  CONSTANT c_end_period         : NATURAL := 10 * c_period_size;
  
  CONSTANT c_usr_fifo_size      : NATURAL := 2 * g_nof_repeat * c_on_period;   -- choose user input FIFO in tb sufficiently large to not overflow before tb_end
  
  CONSTANT c_bsn_max            : UNSIGNED(c_dp_stream_bsn_w-1 DOWNTO 0) := (OTHERS=>'1');
  CONSTANT c_bsn_gap            : UNSIGNED(c_dp_stream_bsn_w-1 DOWNTO 0) := TO_UNSIGNED(ceil_div(c_off_period, c_period_size), c_dp_stream_bsn_w);   -- depends on c_off_period
  
  CONSTANT c_tx_seq_init        : NATURAL := 10;
  
  TYPE t_tb_state IS (s_init, s_bg_enable_sync, s_bg_enable, s_bg_disable, s_user, s_xoff, s_xon);

  SIGNAL tb_end                     : STD_LOGIC := '0';
  SIGNAL rst                        : STD_LOGIC;
  SIGNAL clk                        : STD_LOGIC := '1';
  SIGNAL en_sync                    : STD_LOGIC;
  SIGNAL tb_state                   : t_tb_state;
  SIGNAL prev_tb_state              : t_tb_state;
  SIGNAL random                     : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS=>'0');
  SIGNAL out_siso_ready             : STD_LOGIC;
  SIGNAL out_siso_xon               : STD_LOGIC;
                         
  SIGNAL ram_bg_data_mosi           : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL ram_bg_data_miso           : t_mem_miso := c_mem_miso_rst;
  SIGNAL reg_bg_ctrl_mosi           : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL reg_bg_ctrl_miso           : t_mem_miso := c_mem_miso_rst;
  SIGNAL reg_tx_seq_mosi            : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL reg_tx_seq_miso            : t_mem_miso := c_mem_miso_rst;

  SIGNAL usr_src_in_arr             : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
  SIGNAL usr_src_out_arr            : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);   
  SIGNAL usr_src_out_0_data         : STD_LOGIC_VECTOR(c_buf_dat_w-1 DOWNTO 0);
  SIGNAL usr_fifo_src_in_arr        : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
  SIGNAL usr_fifo_src_out_arr       : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);   
  
  SIGNAL out_siso_arr               : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
  SIGNAL out_sosi_arr               : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);   
  SIGNAL prev_out_sosi_0            : t_dp_sosi;
  SIGNAL out_sosi_0                 : t_dp_sosi;
  SIGNAL out_sosi_0_data            : STD_LOGIC_VECTOR(c_buf_dat_w-1 DOWNTO 0);
  SIGNAL prev_out_sosi_0_data_u0    : STD_LOGIC_VECTOR(c_buf_dat_w-1 DOWNTO 0);
  SIGNAL prev_out_sosi_0_data_u1    : STD_LOGIC_VECTOR(c_buf_dat_w-1 DOWNTO 0);
  
  SIGNAL bg_en                      : STD_LOGIC := '0';
  SIGNAL bg_en_dly                  : STD_LOGIC := '0';
  SIGNAL verify_sosi_arr_en         : STD_LOGIC := '0';
  SIGNAL verify_bsn_en              : STD_LOGIC := '0';
  SIGNAL verify_sync_en             : STD_LOGIC := '0';
  SIGNAL verify_data_en             : STD_LOGIC := '0';
  SIGNAL verify_tx_seq_en           : STD_LOGIC := '0';
  SIGNAL verify_no_bg_buffer_ram_en : STD_LOGIC := '0';
  
BEGIN
  
  clk <= NOT clk OR tb_end AFTER clk_period/2;
  rst <= '1', '0' AFTER clk_period*5;
  
  p_dut_bg_ctrl : PROCESS
  BEGIN
    bg_en <= '0';
    en_sync <= '0';
    out_siso_xon <= '1';  -- Flow control XON
    
    IF c_verify_usr_rflow=TRUE THEN
      tb_state <= s_user;
      proc_common_wait_some_cycles(clk, c_on_period);              -- On time
    ELSE
      -- BG init
      tb_state <= s_init;
      proc_common_wait_until_low(clk, rst);                      -- Wait until reset has finished
      proc_common_wait_some_cycles(clk, 10);                     -- Wait an additional amount of cycles
      proc_mem_mm_bus_wr(1,   c_block_size, clk, reg_bg_ctrl_mosi);  -- Set the nof samples per block
      proc_mem_mm_bus_wr(2,  c_sync_period, clk, reg_bg_ctrl_mosi);  -- Set the nof blocks per sync
      proc_mem_mm_bus_wr(3,     g_gap_size, clk, reg_bg_ctrl_mosi);  -- Set the gapsize
      proc_mem_mm_bus_wr(4,  c_mem_lo_addr, clk, reg_bg_ctrl_mosi);  -- Set the mem low address
      proc_mem_mm_bus_wr(5,  c_mem_hi_addr, clk, reg_bg_ctrl_mosi);  -- Set the mem high address
      proc_mem_mm_bus_wr(6,   c_bsn_offset, clk, reg_bg_ctrl_mosi);  -- Set the lower part of the initial bsn
      proc_mem_mm_bus_wr(7,              0, clk, reg_bg_ctrl_mosi);  -- Set the higher part of the initial bsn
      
      -- BG enable at en_sync pulse
      proc_mem_mm_bus_wr(0,              3, clk, reg_bg_ctrl_mosi);  -- Enable the block generator at en_sync pulse. 
      tb_state <= s_bg_enable_sync;
      proc_common_wait_some_cycles(clk, c_pend_period);          -- Wait an additional amount of cycles
      proc_common_gen_pulse(clk, en_sync);                       -- Issue the en_sync pulse to start the enabled block generator
      bg_en <= '1';
      proc_common_wait_some_cycles(clk, c_on_period);            -- On time
      proc_common_wait_until_high(clk, out_sosi_0.valid);        -- Wait until block generator [0] is active
      proc_common_wait_some_cycles(clk, 20);                     -- Wait an additional amount of cycles  
    
      FOR I IN 0 TO g_nof_repeat-1 LOOP
        -- BG disable
        proc_mem_mm_bus_wr(0,   0, clk, reg_bg_ctrl_mosi);         -- Disable the block generator. 
        tb_state <= s_bg_disable;
        bg_en <= '0';
        proc_common_wait_some_cycles(clk, c_off_period);           -- Off time
        
        -- BG enable
        proc_mem_mm_bus_wr(0,   1, clk, reg_bg_ctrl_mosi);         -- Enable the block generator. 
        tb_state <= s_bg_enable;
        bg_en <= '1';
        proc_common_wait_some_cycles(clk, c_on_period);            -- On time
        proc_common_wait_until_high(clk, out_sosi_0.valid);        -- Wait until block generator [0] is active
        proc_common_wait_some_cycles(clk, 20);                     -- Wait an additional amount of cycles
      END LOOP;
      
      FOR I IN 0 TO g_nof_repeat-1 LOOP
        -- BG flow off
        out_siso_xon <= '0';  -- Flow control XON
        tb_state <= s_xoff;
        bg_en <= '0';
        proc_common_wait_some_cycles(clk, c_off_period);           -- Off time
        
        -- BG flow on
        out_siso_xon <= '1';  -- Flow control XON
        tb_state <= s_xon;
        bg_en <= '1';
        proc_common_wait_some_cycles(clk, c_on_period);            -- On time
      END LOOP;
    END IF;
    
    proc_common_wait_some_cycles(clk, c_end_period);             -- Some extra time
    tb_end <= '1';
    WAIT;
  END PROCESS;
      
  p_dut_tx_seq_ctrl : PROCESS
  BEGIN
    proc_common_wait_until_low(clk, rst);    -- Wait until reset has finished
    proc_common_wait_some_cycles(clk, 10);   -- Wait an additional amount of cycles
    proc_mem_mm_bus_wr(1, c_tx_seq_init, clk, reg_tx_seq_mosi);                 -- Init Tx sequence
    proc_common_wait_some_cycles(clk, 10);   -- Wait for some latency
    proc_mem_mm_bus_wr(0, c_diag_seq_tx_reg_en_cntr, clk, reg_tx_seq_mosi);     -- Enable Tx CNTR sequence
    WAIT;
  END PROCESS;
  
  
  -------------------------------------------------
  -- User input modelled by another BG
  -------------------------------------------------
  u_user : ENTITY work.mms_diag_block_gen
  GENERIC MAP (    
    g_nof_streams        => g_nof_streams, 
    g_buf_dat_w          => c_buf_dat_w,       
    g_buf_addr_w         => c_buf_addr_w,  
    g_file_name_prefix   => c_file_name_prefix,
    g_diag_block_gen_rst => c_diag_block_gen_enabled  -- user BG is default enabled
  )
  PORT MAP (
    -- System
    mm_rst           => rst, 
    mm_clk           => clk, 
    dp_rst           => rst, 
    dp_clk           => clk, 
    en_sync          => '0',
    -- ST interface
    out_siso_arr     => usr_src_in_arr,
    out_sosi_arr     => usr_src_out_arr
  );

  -- Use sufficiently large FIFO to provide siso.ready flow control to the user input
  no_user_fifo : IF c_use_user_fifo=FALSE GENERATE
    usr_src_in_arr       <= usr_fifo_src_in_arr;
    usr_fifo_src_out_arr <= usr_src_out_arr;
  END GENERATE;
  
  gen_user_fifo : IF c_use_user_fifo=TRUE GENERATE
    gen_nof_streams : FOR I IN 0 TO g_nof_streams-1 GENERATE
      u_user_fifo : ENTITY dp_lib.dp_fifo_sc
      GENERIC MAP (
        g_data_w         => c_buf_dat_w,
        g_bsn_w          => c_dp_stream_bsn_w,
        g_empty_w        => c_dp_stream_empty_w,
        g_channel_w      => c_dp_stream_channel_w,
        g_error_w        => c_dp_stream_error_w,
        g_use_bsn        => TRUE,
        g_use_empty      => TRUE,
        g_use_channel    => TRUE,
        g_use_error      => TRUE,
        g_use_sync       => TRUE,
        g_use_ctrl       => TRUE,
        g_use_complex    => FALSE,
        g_fifo_size      => c_usr_fifo_size,
        g_fifo_af_margin => 4,
        g_fifo_rl        => 1
      )
      PORT MAP (
        rst         => rst,
        clk         => clk,
        -- Monitor FIFO filling
        wr_ful      => OPEN,
        usedw       => OPEN,
        rd_emp      => OPEN,
        -- ST sink
        snk_out     => usr_src_in_arr(I),
        snk_in      => usr_src_out_arr(I),
        -- ST source
        src_in      => usr_fifo_src_in_arr(I), 
        src_out     => usr_fifo_src_out_arr(I)
      );
    END GENERATE;
  END GENERATE;
  
  ------------------------------------------------------------------------------
  -- STREAM CONTROL
  ------------------------------------------------------------------------------
  
  random <= func_common_random(random) WHEN rising_edge(clk);
  
  out_siso_ready <= '1'                 WHEN g_flow_control_verify=e_active  ELSE
                    random(random'HIGH) WHEN g_flow_control_verify=e_random  ELSE
                    '1';

  p_out_siso_arr : PROCESS(out_siso_ready, out_siso_xon)
  BEGIN
    FOR I IN 0 TO g_nof_streams-1 LOOP
      out_siso_arr(I).ready <= '1';
      out_siso_arr(I).xon <= out_siso_xon;
    END LOOP;
    IF c_verify_usr_rflow=TRUE THEN
      out_siso_arr(0).ready <= out_siso_ready;  -- only verify flow control via stream 0
    END IF;
  END PROCESS;
                     
  -------------------------------------------------
  -- Device under test
  -------------------------------------------------
  u_dut : ENTITY work.mms_diag_block_gen
  GENERIC MAP (    
    -- Generate configurations
    g_use_usr_input      => g_use_usr_input,
    g_use_bg             => g_use_bg,
    g_use_tx_seq         => g_use_tx_seq,
    -- General
    g_nof_streams        => g_nof_streams,
    -- BG settings
    g_use_bg_buffer_ram  => g_use_bg_buffer_ram,
    g_buf_dat_w          => c_buf_dat_w,       
    g_buf_addr_w         => c_buf_addr_w,  
    g_file_name_prefix   => c_file_name_prefix,     
    g_diag_block_gen_rst => c_diag_block_gen_rst,  -- user BG is default disabled, MM controlled by p_dut_bg_ctrl
    -- User input multiplexer option
    g_usr_bypass_xonoff  => g_usr_bypass_xonoff,
    -- Tx_seq
    g_seq_dat_w          => c_buf_dat_w  -- >= 1, test sequence data width. Choose g_seq_dat_w <= g_buf_dat_w
  )
  PORT MAP (
    -- System
    mm_rst           => rst, 
    mm_clk           => clk, 
    dp_rst           => rst, 
    dp_clk           => clk, 
    en_sync          => en_sync,
    -- MM interface
    ram_bg_data_mosi => ram_bg_data_mosi,
    ram_bg_data_miso => ram_bg_data_miso,
    reg_bg_ctrl_mosi => reg_bg_ctrl_mosi,
    reg_bg_ctrl_miso => reg_bg_ctrl_miso,
    reg_tx_seq_mosi  => reg_tx_seq_mosi,
    reg_tx_seq_miso  => reg_tx_seq_miso,
    -- ST interface
    usr_siso_arr     => usr_fifo_src_in_arr,
    usr_sosi_arr     => usr_fifo_src_out_arr,
    out_siso_arr     => out_siso_arr,
    out_sosi_arr     => out_sosi_arr
  );

  -------------------------------------------------
  -- Verification
  -------------------------------------------------
  
  prev_tb_state <= tb_state WHEN rising_edge(clk);
  
  usr_src_out_0_data <= usr_src_out_arr(0).data(c_buf_dat_w-1 DOWNTO 0);
  
  out_sosi_0      <= out_sosi_arr(0);  -- use BG[0] as reference
  out_sosi_0_data <= out_sosi_0.data(c_buf_dat_w-1 DOWNTO 0);
  
  -- Verify : Equal data in all g_nof_streams
  verify_sosi_arr_en <= '1' WHEN bg_en_dly='1';
    
  p_compare_out_sosi_arr : PROCESS(clk)
  BEGIN
    IF rising_edge(clk) THEN
      IF verify_sosi_arr_en='1' THEN
        FOR I IN 0 TO g_nof_streams-1 LOOP
          -- All outputs have the same framing
          ASSERT out_sosi_0.sync =out_sosi_arr(I).sync  REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).sync";
          ASSERT out_sosi_0.bsn  =out_sosi_arr(I).bsn   REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).bsn";
          ASSERT out_sosi_0.im   =out_sosi_arr(I).im    REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).im";
          ASSERT out_sosi_0.valid=out_sosi_arr(I).valid REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).valid";
          ASSERT out_sosi_0.sop  =out_sosi_arr(I).sop   REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).sop";
          ASSERT out_sosi_0.eop  =out_sosi_arr(I).eop   REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).eop";
          
          IF g_use_bg_buffer_ram=TRUE THEN
            -- Only the re part differs per output given the BG buffer data from c_file_name_prefix = "data/bf_in_data"
            ASSERT TO_UINT(out_sosi_0.re)=TO_UINT(out_sosi_arr(I).re)-I*4 REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).re";
          ELSE
            ASSERT out_sosi_0.im=out_sosi_arr(I).im REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).im";
          END IF;
        END LOOP;
      END IF;
    END IF;
  END PROCESS;

  -- Verify : BSN
  p_verify_bsn_en : PROCESS
  BEGIN
    proc_common_wait_until_evt(clk, bg_en);
    verify_bsn_en <= '0';
    proc_common_wait_some_cycles(clk, 2*c_period_size);
    verify_bsn_en <= '1';
    WAIT UNTIL rising_edge(clk);
  END PROCESS;
  
  proc_dp_verify_data("BSN", c_bsn_max, c_bsn_gap, clk, verify_bsn_en, out_sosi_0.sop, out_sosi_0.bsn, prev_out_sosi_0.bsn);  -- verify both BG and user
  
  -- Verify : sync
  bg_en_dly <= bg_en'DELAYED(5*clk_period);
  verify_sync_en <= verify_bsn_en AND bg_en_dly WHEN g_use_bg=TRUE ELSE '0';  -- only verify BG sync interval (ignore user sync interval)
  
  proc_dp_verify_sync(c_sync_period, c_sync_offset, clk, verify_sync_en, out_sosi_0.sync, out_sosi_0.sop, out_sosi_0.bsn);
  
  -- Verify : Tx counter data
  -- . g_use_tx_seq=TRUE
  -- . g_use_bg_buffer_ram=FALSE
  p_verify_data_en : PROCESS
  BEGIN
    verify_data_en <= '0';
    IF g_use_usr_input=TRUE THEN
      proc_common_wait_until_high(clk, usr_fifo_src_out_arr(0).valid);
    END IF;
    IF g_use_bg=TRUE THEN
      proc_common_wait_until_evt(clk, bg_en);
    END IF;
    proc_common_wait_some_cycles(clk, 2*c_period_size);
    verify_data_en <= '1';
    
    WHILE TRUE LOOP
      IF tb_state=s_xoff THEN
        verify_data_en <= '0';
      ELSIF prev_tb_state=s_xoff THEN
        proc_common_wait_some_cycles(clk, 2*c_period_size);
        verify_data_en <= '1';
      END IF;
      WAIT UNTIL rising_edge(clk);
    END LOOP;
  END PROCESS;
  
  verify_tx_seq_en           <= '1' WHEN verify_data_en='1' AND g_use_tx_seq=TRUE         ELSE '0';
  verify_no_bg_buffer_ram_en <= '1' WHEN verify_data_en='1' AND g_use_bg_buffer_ram=FALSE ELSE '0';
  
  proc_dp_verify_data("Tx sequence counter data",                   0, 0, clk, verify_tx_seq_en,           out_sosi_0.valid, out_sosi_0_data, prev_out_sosi_0_data_u0);
  proc_dp_verify_data("No BG buffer RAM counter data", c_block_size-1, 0, clk, verify_no_bg_buffer_ram_en, out_sosi_0.valid, out_sosi_0_data, prev_out_sosi_0_data_u1);
   
  -- Verify : User input bypass all
  p_compare_usr_bypass_all : PROCESS(clk)
  BEGIN
    IF g_use_usr_input=TRUE AND g_use_bg=FALSE AND g_use_tx_seq=FALSE THEN
      IF rising_edge(clk) THEN
        ASSERT out_sosi_arr       =usr_fifo_src_out_arr REPORT "tb_mms_diag_block_gen: out_sosi_arr   /= usr_src_out_arr";
        ASSERT usr_fifo_src_in_arr=out_siso_arr         REPORT "tb_mms_diag_block_gen: usr_src_in_arr /= out_siso_arr";
      END IF;
    END IF;
  END PROCESS;
  
END tb;
