-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose:
-- . Development version that will eventually replace dp_offload_tx.
-- Description:
-- . From each incoming streams, dp_offload_tx allows the user to control:
--   . The number of words per incoming frame that is selected 
--   . The numer of blocks of these selected words that are put into one new frame
-- . The above control settings can be done dynaically via the MM interface. If the MM
--   interface is not used, the default settings passed via g_hdr_field_arr are used.
-- . NOTE: dp_frame_rd DOES NOT SUPPORT 1 WORD FRAMES due to its separate s_sof and
--   s_eof states.
-- Remarks:
-- . Note that an output (fill) FIFO may be needed if the packets have gaps in them!
--   The Ethernet cores require this so an EOP-triggered FIFO may added to them.

LIBRARY IEEE, common_lib, work, mm_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE common_lib.common_field_pkg.ALL;

ENTITY dp_offload_tx IS
  GENERIC (
    g_nof_streams               : NATURAL;
    g_data_w                    : NATURAL;
    g_use_complex               : BOOLEAN; -- TRUE uses re(0..g_data_w/2 -1) & im(0..g_data_w/2-1) as input instead of data(0..g_data_w-1).
    g_max_nof_words_per_block   : NATURAL;
    g_def_nof_words_per_block   : NATURAL;
    g_max_nof_blocks_per_packet : NATURAL;
    g_def_nof_blocks_per_packet : NATURAL;
    g_hdr_field_arr             : t_common_field_arr;
    g_hdr_field_sel             : STD_LOGIC_VECTOR;
    g_use_post_split_fifo       : BOOLEAN := FALSE
  ); 
  PORT (
    mm_rst               : IN  STD_LOGIC;
    mm_clk               : IN  STD_LOGIC; 
    
    dp_rst               : IN  STD_LOGIC;
    dp_clk               : IN  STD_LOGIC;  

    reg_mosi             : IN  t_mem_mosi;
    reg_miso             : OUT t_mem_miso;

    reg_hdr_dat_mosi     : IN  t_mem_mosi;
    reg_hdr_dat_miso     : OUT t_mem_miso;

    snk_in_arr           : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0); 
    snk_out_arr          : OUT t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);              

    src_out_arr          : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0); 
    src_in_arr           : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);

    hdr_fields_in_arr    : IN  t_slv_1024_arr(g_nof_streams-1 DOWNTO 0) -- hdr_fields_in_arr(i) is considered valid @ snk_in_arr(i).sop
  );
END dp_offload_tx;


ARCHITECTURE str OF dp_offload_tx IS 

  CONSTANT c_dp_split_val_latency        : NATURAL := 1;
  CONSTANT c_dp_packet_merge_val_latency : NATURAL := 2;
  CONSTANT c_nof_header_words            : NATURAL := field_slv_len(g_hdr_field_arr) / g_data_w;
  CONSTANT c_dp_field_blk_snk_data_w     : NATURAL := field_slv_out_len(field_arr_set_mode(g_hdr_field_arr , "RW"));
  CONSTANT c_dp_field_blk_src_data_w     : NATURAL := g_data_w;

  SIGNAL dp_split_snk_in_arr          : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL dp_split_snk_out_arr         : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
 
  SIGNAL dp_split_src_out_2arr        : t_dp_sosi_2arr_2(g_nof_streams-1 DOWNTO 0);
  SIGNAL dp_split_src_in_2arr         : t_dp_siso_2arr_2(g_nof_streams-1 DOWNTO 0);

  SIGNAL dp_packet_merge_snk_in_arr   : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL dp_packet_merge_snk_out_arr  : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);

  SIGNAL dp_packet_merge_src_out_arr  : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL dp_packet_merge_src_in_arr   : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);

  SIGNAL dp_field_blk_snk_in_arr      : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL dp_field_blk_snk_out_arr     : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);

  SIGNAL dp_concat_snk_in_2arr        : t_dp_sosi_2arr_2(g_nof_streams-1 DOWNTO 0);
  SIGNAL dp_concat_snk_out_2arr       : t_dp_siso_2arr_2(g_nof_streams-1 DOWNTO 0);

  -- MM control
  CONSTANT c_nof_ctrl_fields : NATURAL := 2;
  CONSTANT c_ctrl_field_arr  : t_common_field_arr(c_nof_ctrl_fields-1 DOWNTO 0) := ( ( field_name_pad("nof_words_per_block"  ), "RW", ceil_log2(g_max_nof_words_per_block+1  ), field_default(g_def_nof_words_per_block) ),
                                                                                     ( field_name_pad("nof_blocks_per_packet"), "RW", ceil_log2(g_max_nof_blocks_per_packet+1), field_default(g_def_nof_blocks_per_packet) ));

  TYPE t_ctrl_fields_out_arr IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(field_slv_out_len(c_ctrl_field_arr)-1 DOWNTO 0);
  SIGNAL ctrl_fields_out_arr : t_ctrl_fields_out_arr;

  TYPE t_nof_words_per_block_arr IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(ceil_log2(g_max_nof_words_per_block+1)-1 DOWNTO 0);
  SIGNAL nof_words_per_block_arr : t_nof_words_per_block_arr; 

  TYPE t_nof_blocks_per_packet_arr IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(ceil_log2(g_max_nof_blocks_per_packet+1)-1 DOWNTO 0);
  SIGNAL nof_blocks_per_packet_arr : t_nof_blocks_per_packet_arr; 

  SIGNAL reg_mosi_arr : t_mem_mosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL reg_miso_arr : t_mem_miso_arr(g_nof_streams-1 DOWNTO 0);

  SIGNAL reg_hdr_dat_mosi_arr : t_mem_mosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL reg_hdr_dat_miso_arr : t_mem_miso_arr(g_nof_streams-1 DOWNTO 0); 

  SIGNAL dp_pipeline_hdr_fields_snk_in_arr  : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL dp_pipeline_hdr_fields_snk_out_arr : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);

  SIGNAL dp_pipeline_hdr_fields_src_out_arr : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL dp_pipeline_hdr_fields_src_in_arr  : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
     
BEGIN

  ---------------------------------------------------------------------------------------
  -- Input: use complex or data field 
  ---------------------------------------------------------------------------------------
  gen_wires_complex : IF g_use_complex = TRUE GENERATE
    p_wires_complex : PROCESS(snk_in_arr)
    BEGIN
      FOR i IN 0 TO g_nof_streams-1 LOOP
        dp_split_snk_in_arr(i) <= snk_in_arr(i);
        dp_split_snk_in_arr(i).data(g_data_w-1 DOWNTO 0) <= snk_in_arr(i).re(g_data_w/2-1 DOWNTO 0) & snk_in_arr(i).im(g_data_w/2-1 DOWNTO 0);
      END LOOP;
    END PROCESS;
  END GENERATE;

  gen_wires_data : IF g_use_complex = FALSE GENERATE
    dp_split_snk_in_arr <= snk_in_arr;
  END GENERATE;
   
  ---------------------------------------------------------------------------------------
  -- Split nof_words_per_block
  ---------------------------------------------------------------------------------------
  snk_out_arr <= dp_split_snk_out_arr;

  gen_dp_split : FOR i IN 0 TO g_nof_streams-1 GENERATE
    u_dp_split : ENTITY work.dp_split
    GENERIC MAP (
      g_data_w         => g_data_w,
      g_symbol_w       => g_data_w,
      g_nof_symbols    => g_max_nof_words_per_block
    )
    PORT MAP (
      rst             => dp_rst,
      clk             => dp_clk,

      nof_symbols     => nof_words_per_block_arr(i),

      snk_out         => dp_split_snk_out_arr(i),
      snk_in          => dp_split_snk_in_arr(i),

      src_in_arr      => dp_split_src_in_2arr(i), 
      src_out_arr     => dp_split_src_out_2arr(i)
    );

    dp_split_src_in_2arr(i)(0) <= c_dp_siso_rdy; -- Always ready to throw away the tail

  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- Use a post-split FIFO when the source has no flow control and its inter-block gap
  -- is shorter than required for the header to be inserted.
  ---------------------------------------------------------------------------------------
  gen_dp_fifo_sc : FOR i IN 0 TO sel_a_b(g_use_post_split_fifo, g_nof_streams-1, -1) GENERATE
    u_dp_fifo_sc : ENTITY work.dp_fifo_sc
    GENERIC MAP (
      g_data_w    => g_data_w,
      g_fifo_size => g_max_nof_words_per_block 
     )
    PORT MAP (
      rst        => dp_rst,
      clk        => dp_clk,
  
      snk_in     => dp_split_src_out_2arr(i)(1),
      snk_out    => dp_split_src_in_2arr(i)(1),
  
      src_out    => dp_packet_merge_snk_in_arr(i),
      src_in     => dp_packet_merge_snk_out_arr(i)
    );
  END GENERATE;

  no_dp_fifo_sc : FOR i IN 0 TO sel_a_b(g_use_post_split_fifo, -1, g_nof_streams-1) GENERATE
    dp_packet_merge_snk_in_arr(i) <= dp_split_src_out_2arr(i)(1);
    dp_split_src_in_2arr(i)(1)    <= dp_packet_merge_snk_out_arr(i);
  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- Merge nof_blocks_per_packet
  ---------------------------------------------------------------------------------------
  gen_dp_packet_merge : FOR i IN 0 TO g_nof_streams-1 GENERATE
    u_dp_packet_merge : ENTITY work.dp_packet_merge
    GENERIC MAP (
      g_nof_pkt => g_max_nof_blocks_per_packet
    )
    PORT MAP (
      rst         => dp_rst,
      clk         => dp_clk,
  
      nof_pkt     => nof_blocks_per_packet_arr(i),
 
      snk_out     => dp_packet_merge_snk_out_arr(i),
      snk_in      => dp_packet_merge_snk_in_arr(i),
  
      src_in      => dp_packet_merge_src_in_arr(i),
      src_out     => dp_packet_merge_src_out_arr(i)
    );

    dp_concat_snk_in_2arr(i)(0) <= dp_packet_merge_src_out_arr(i);
    dp_packet_merge_src_in_arr(i) <= dp_concat_snk_out_2arr(i)(0);

  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- Pipeline the header SLV so it lines up with dp_packet_merge_src_out_arr(i).sop  
  ---------------------------------------------------------------------------------------
  gen_dp_pipeline_hdr_fields_in_arr : FOR i IN 0 TO g_nof_streams-1 GENERATE

    -- FIXME - Put this DP bus on the entity instead of hdr_fields_in_arr.
    dp_pipeline_hdr_fields_snk_in_arr(i).data(field_slv_len(g_hdr_field_arr)-1 DOWNTO 0) <= hdr_fields_in_arr(i)(field_slv_len(g_hdr_field_arr)-1 DOWNTO 0);
    dp_pipeline_hdr_fields_snk_in_arr(i).valid <= snk_in_arr(i).sop;

    u_dp_pipeline : ENTITY work.dp_pipeline
    GENERIC MAP (
      g_pipeline  => c_dp_split_val_latency + c_dp_packet_merge_val_latency
    )
    PORT MAP (
      rst     => dp_rst,
      clk     => dp_clk,

      snk_in  => dp_pipeline_hdr_fields_snk_in_arr(i),
      snk_out => OPEN, --dp_pipeline_hdr_fields_snk_out_arr(i), -- snk_in is only valid at the SOP so we won't need this

      src_out => dp_pipeline_hdr_fields_src_out_arr(i),
      src_in  => dp_pipeline_hdr_fields_src_in_arr(i)
    );
  END GENERATE;
   
  ---------------------------------------------------------------------------------------
  -- Create header block & concatenate header to offload stream.
  -- . hdr_fields_in_arr(i) is clocked in at snk_in_arr(i).sop after which it can be
  --   read out by dp_concat when it is ready for it.
  ---------------------------------------------------------------------------------------
  gen_dp_field_blk : FOR i IN 0 TO g_nof_streams-1 GENERATE

    -- Flow control for the header source. Should not be needed.
    dp_pipeline_hdr_fields_src_in_arr(i) <= dp_field_blk_snk_out_arr(i);
    dp_field_blk_snk_in_arr(i) <= dp_pipeline_hdr_fields_src_out_arr(i);

    u_dp_field_blk : ENTITY work.dp_field_blk
    GENERIC MAP (
      g_field_arr   => field_arr_set_mode(g_hdr_field_arr , "RW"),
      g_field_sel   => g_hdr_field_sel,
      g_snk_data_w  => c_dp_field_blk_snk_data_w,
      g_src_data_w  => c_dp_field_blk_src_data_w
    )
    PORT MAP (
      dp_rst       => dp_rst,
      dp_clk       => dp_clk,

      mm_rst       => mm_rst,
      mm_clk       => mm_clk,
  
      snk_in       => dp_field_blk_snk_in_arr(i), -- Valid for only one cycle.
      snk_out      => dp_field_blk_snk_out_arr(i),
 
      src_in       => dp_concat_snk_out_2arr(i)(1),
      src_out      => dp_concat_snk_in_2arr(i)(1),

      reg_slv_mosi => reg_hdr_dat_mosi_arr(i),
      reg_slv_miso => reg_hdr_dat_miso_arr(i)
    );
  END GENERATE;

  gen_dp_concat : FOR i IN 0 TO g_nof_streams-1 GENERATE
    u_dp_concat : ENTITY work.dp_concat
    GENERIC MAP (
      g_data_w    => g_data_w,
      g_symbol_w  => c_byte_w 
    )
    PORT MAP (
      rst         => dp_rst,
      clk         => dp_clk,

      snk_out_arr => dp_concat_snk_out_2arr(i),
      snk_in_arr  => dp_concat_snk_in_2arr(i),

      src_in      => src_in_arr(i),
      src_out     => src_out_arr(i)
    ); 
  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- MM control & monitoring
  ---------------------------------------------------------------------------------------
  gen_mm_fields: FOR i IN 0 TO g_nof_streams-1 GENERATE
    u_mm_fields: ENTITY mm_lib.mm_fields
    GENERIC MAP(
      g_field_arr => c_ctrl_field_arr
    )
    PORT MAP (
      mm_clk  => mm_clk,
      mm_rst  => mm_rst,
  
      mm_mosi => reg_mosi_arr(i),
      mm_miso => reg_miso_arr(i),
      
      slv_clk => dp_clk,
      slv_rst => dp_rst, 
  
      slv_out => ctrl_fields_out_arr(i)
    );

    nof_words_per_block_arr(i)   <= ctrl_fields_out_arr(i)(field_hi(c_ctrl_field_arr, "nof_words_per_block"  ) DOWNTO field_lo(c_ctrl_field_arr, "nof_words_per_block"  ));
    nof_blocks_per_packet_arr(i) <= ctrl_fields_out_arr(i)(field_hi(c_ctrl_field_arr, "nof_blocks_per_packet") DOWNTO field_lo(c_ctrl_field_arr, "nof_blocks_per_packet"));

  END GENERATE;

  u_common_mem_mux_ctrl : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_streams,
    g_mult_addr_w => ceil_log2(field_nof_words(c_ctrl_field_arr, c_word_w))
  )
  PORT MAP (
    mosi     => reg_mosi,
    miso     => reg_miso,
    mosi_arr => reg_mosi_arr,
    miso_arr => reg_miso_arr
  );

  u_common_mem_mux_hdr_dat : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_streams,
    g_mult_addr_w => ceil_log2(field_nof_words(g_hdr_field_arr, c_word_w))
  )
  PORT MAP (
    mosi     => reg_hdr_dat_mosi,
    miso     => reg_hdr_dat_miso,
    mosi_arr => reg_hdr_dat_mosi_arr,
    miso_arr => reg_hdr_dat_miso_arr
  );

END str;
