--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose: DP wrapper for common_reinterleave
-- Description:
-- Remarks:
-- . Requirements: 
--   * All input streams must be synchronous;
--   * g_inter_block_size >= g_deint_block_size;
--   * Correct forwarding of SOP, EOP and Sync (g_use_ctrl=TRUE) requires that:
--    . g_nof_in = g_nof_out;
--    . No gaps between SOP and EOP;
--    . g_deint_block_size = g_block_size out
--    . Input block length (!= g_deint_block_size) / nof_out / g_block_size_x = integer. 
--      If it is not an integer number, the input blocks cannot be evenly distributed over
--      the outputs with the chosen block_size_x.

ENTITY dp_reinterleave IS
  GENERIC (
    g_dat_w          : NATURAL;
    g_nof_in         : NATURAL;
    g_deint_block_size  : NATURAL; -- Inputs streams will be de-interleaved using this block size. Should match block_size_out when g_use_ctrl=TRUE
    g_nof_out        : NATURAL;
    g_inter_block_size : NATURAL; -- Outputs streams will be interleaved using this block size. Should match block_size_in when g_use_ctrl=TRUE
    g_use_ctrl       : BOOLEAN := TRUE;
    g_use_complex    : BOOLEAN;
    g_align_out  : BOOLEAN := FALSE
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;

    snk_in_arr  : IN  t_dp_sosi_arr(g_nof_in-1 DOWNTO 0);
    src_out_arr : OUT t_dp_sosi_arr(g_nof_out-1 DOWNTO 0)
  );
END dp_reinterleave;


ARCHITECTURE wrap OF dp_reinterleave IS

  CONSTANT c_complex_w : NATURAL := g_dat_w/2;

  -- common_interleave latency: multiplexer input register stage
  CONSTANT c_inter_mux_latency : NATURAL := 1;

  -- common_interleave latency: common_blockreg latencies: see common_blockreg.vhd for details
  CONSTANT c_common_blockreg_latency       : NATURAL := 1;
  CONSTANT c_common_blockreg_valid_latency : NATURAL := g_inter_block_size*g_nof_out - (g_deint_block_size*g_nof_out - g_deint_block_size);

  -- Arrays flattened into SLVs
  SIGNAL in_dat        : STD_LOGIC_VECTOR(g_nof_in*g_dat_w-1 DOWNTO 0);

  SIGNAL out_dat       : STD_LOGIC_VECTOR(g_nof_out*g_dat_w-1 DOWNTO 0);
  SIGNAL out_val       : STD_LOGIC_VECTOR(g_nof_out-1 DOWNTO 0);

BEGIN

  -----------------------------------------------------------------------------
  -- Map input sosi_arr to SLV
  -----------------------------------------------------------------------------
  gen_wires_in : FOR i IN 0 TO g_nof_in-1 GENERATE 
    gen_sosi_dat_in: IF g_use_complex = FALSE GENERATE
      -----------------------------------------------------------------------------
      -- Forward SOSI data field
      -----------------------------------------------------------------------------
      in_dat( i*g_dat_w+g_dat_w -1 DOWNTO i*g_dat_w) <= snk_in_arr(i).data(g_dat_w-1 DOWNTO 0);
    END GENERATE;

    gen_sosi_compl_in: IF g_use_complex = TRUE GENERATE
      -----------------------------------------------------------------------------
      -- Forward concatenated im & re fields
      -----------------------------------------------------------------------------
      in_dat( i*g_dat_w+g_dat_w -1 DOWNTO i*g_dat_w) <= snk_in_arr(i).im(c_complex_w-1 DOWNTO 0) & snk_in_arr(i).re(c_complex_w-1 DOWNTO 0);
    END GENERATE;
  END GENERATE;

  u_reinterleave : ENTITY common_lib.common_reinterleave
  GENERIC MAP (
    g_nof_in         => g_nof_in,
    g_deint_block_size  => g_deint_block_size,
    g_nof_out        => g_nof_out,
    g_inter_block_size => g_inter_block_size,
    g_dat_w          => g_dat_w,
    g_align_out => g_align_out
  )
  PORT MAP (
    rst        => rst,
    clk        => clk,
    
    in_dat     => in_dat,
    in_val     => snk_in_arr(0).valid, -- All input streams should be synchronous
    
    out_dat    => out_dat,
    out_val    => out_val
  );

  -----------------------------------------------------------------------------
  -- Map output SLV to sosi_arr
  -----------------------------------------------------------------------------
  gen_wires_out : FOR i IN 0 TO g_nof_out-1 GENERATE
    gen_sosi_dat_out: IF g_use_complex = FALSE GENERATE
      -----------------------------------------------------------------------------
      -- Forward SOSI data field
      -----------------------------------------------------------------------------
      src_out_arr(i).data(g_dat_w-1 DOWNTO 0) <= out_dat( i*g_dat_w+g_dat_w -1 DOWNTO i*g_dat_w);
    END GENERATE;
    
    gen_sosi_compl_out: IF g_use_complex = TRUE GENERATE
      -----------------------------------------------------------------------------
      -- Forward concatenated im & re fields
      -----------------------------------------------------------------------------
      src_out_arr(i).im <= RESIZE_DP_DSP_DATA(out_dat( i*g_dat_w + g_dat_w     -1 DOWNTO i*g_dat_w + c_complex_w));
      src_out_arr(i).re <= RESIZE_DP_DSP_DATA(out_dat( i*g_dat_w + c_complex_w -1 DOWNTO i*g_dat_w));
    END GENERATE;

    src_out_arr(i).valid <= out_val(i);
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Forward SOP, EOP and Sync of snk_in(0) to all outputs, with the correct 
  -- delays. Don't compensate for asynchronous deinterleaver output streams
  -- when g_align_out is true.
  -----------------------------------------------------------------------------
  gen_ctrl : if g_use_ctrl=TRUE GENERATE
    gen_pipe_sop: FOR i IN 0 TO g_nof_out-1 GENERATE 
      u_shiftreg_sop : ENTITY common_lib.common_shiftreg
      GENERIC MAP (
        g_pipeline  => c_common_blockreg_latency + c_inter_mux_latency,
        g_nof_dat   => 1 + c_common_blockreg_valid_latency + sel_a_b(g_align_out, (g_nof_out-1)*g_deint_block_size, (i*g_deint_block_size)),
        g_dat_w     => 1 
      )
      PORT MAP (
        rst          => rst,
        clk          => clk,
        
        in_dat       => slv(snk_in_arr(0).sop),        
        sl(out_dat)  => src_out_arr(i).sop
      );
 
      u_shiftreg_eop : ENTITY common_lib.common_shiftreg
      GENERIC MAP (
        g_pipeline  => c_common_blockreg_latency + c_inter_mux_latency,
        g_nof_dat   => 1 + c_common_blockreg_valid_latency + sel_a_b(g_align_out, (g_nof_out-1)*g_deint_block_size, (i*g_deint_block_size)),
        g_dat_w     => 1,
        g_flush_en  => TRUE -- Flush out the EOP immediately
      )
      PORT MAP (
        rst          => rst,
        clk          => clk,
        
        in_dat       => slv(snk_in_arr(0).eop),
        in_eop       => snk_in_arr(0).eop, 
        sl(out_dat)  => src_out_arr(i).eop  
       );
 
      u_shiftreg_sync : ENTITY common_lib.common_shiftreg
      GENERIC MAP (
        g_pipeline  => c_common_blockreg_latency + c_inter_mux_latency,
        g_nof_dat   => 1 + c_common_blockreg_valid_latency + sel_a_b(g_align_out, (g_nof_out-1)*g_deint_block_size, (i*g_deint_block_size)), 
        g_dat_w     => 1 
      )
      PORT MAP (
        rst          => rst,
        clk          => clk,
        
        in_dat       => slv(snk_in_arr(0).sync),       
        sl(out_dat)  => src_out_arr(i).sync
      );

      u_shiftreg_bsn : ENTITY common_lib.common_shiftreg
      GENERIC MAP (
        g_pipeline  => c_common_blockreg_latency + c_inter_mux_latency,
        g_nof_dat   => 1 + c_common_blockreg_valid_latency + sel_a_b(g_align_out, (g_nof_out-1)*g_deint_block_size, (i*g_deint_block_size)), 
        g_dat_w     => c_dp_stream_bsn_w
      )
      PORT MAP (
        rst          => rst,
        clk          => clk,
        
        in_dat       => snk_in_arr(0).bsn,       
        out_dat      => src_out_arr(i).bsn
      );
  
    END GENERATE;
  END GENERATE;

END wrap;
