-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_lfsr_sequences_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE work.tb_dp_pkg.ALL;

-- Purpose: Verify dp_latency_fifo
-- Description:
--   p_stimuli --> dp_latency_fifo --> verify
-- Usage:
-- > as 10
-- > run -all  -- signal tb_end will stop the simulation by stopping the clk

ENTITY tb_dp_latency_fifo IS
  GENERIC (
    g_bypass                : BOOLEAN := FALSE;  -- for clarity rather use g_bypass=TRUE than equivalent g_fifo_size=0
    g_input_rl              : NATURAL := 1;      -- input ready latency
    g_output_rl             : NATURAL := 1;      -- output ready latency
    g_fifo_size             : NATURAL := 1;
    g_nof_symbols_per_block : NATURAL := 51;
    g_nof_symbols_per_data  : NATURAL := 4;
    g_in_en                 : t_dp_flow_control_enum := e_pulse;  -- always e_active, e_random or e_pulse flow control
    g_out_ready             : t_dp_flow_control_enum := e_random   -- always e_active, e_random or e_pulse flow control
  );
END tb_dp_latency_fifo;


ARCHITECTURE tb OF tb_dp_latency_fifo IS

  CONSTANT c_nof_data_per_block       : NATURAL := ceil_div(g_nof_symbols_per_block, g_nof_symbols_per_data);
  
  -- tb default
  CONSTANT c_pulse_0_active           : NATURAL := c_nof_data_per_block;
  CONSTANT c_pulse_0_gap              : NATURAL := 10;
  CONSTANT c_pulse_0_period           : NATURAL := c_nof_data_per_block + c_pulse_0_gap;
  
  CONSTANT c_pulse_1_active           : NATURAL := 1;
  CONSTANT c_pulse_1_period           : NATURAL := 7;
  
  -- tb specific
  CONSTANT c_nof_repeat               : NATURAL := 20;
  CONSTANT c_symbol_w                 : NATURAL := 8;
  CONSTANT c_data_w                   : NATURAL := g_nof_symbols_per_data * c_symbol_w;
  CONSTANT c_symbol_init              : NATURAL := 0;
  CONSTANT c_symbol_mod               : INTEGER := 2**c_symbol_w;  -- used to avoid TO_UVEC warning for smaller c_symbol_w : "NUMERIC_STD.TO_UNSIGNED: vector truncated"
  
  -- tb default
  SIGNAL tb_end               : STD_LOGIC := '0';
  SIGNAL clk                  : STD_LOGIC := '1';
  SIGNAL rst                  : STD_LOGIC := '1';

  SIGNAL random_0             : STD_LOGIC_VECTOR(14 DOWNTO 0) := (OTHERS=>'0');  -- use different lengths to have different random sequences
  SIGNAL random_1             : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS=>'0');  -- use different lengths to have different random sequences
  SIGNAL pulse_0              : STD_LOGIC;
  SIGNAL pulse_1              : STD_LOGIC;
  SIGNAL pulse_en             : STD_LOGIC := '1';
  
  SIGNAL in_en                : STD_LOGIC := '1';
  
  -- tb verify
  SIGNAL verify_en            : STD_LOGIC := '0';
  SIGNAL verify_done          : STD_LOGIC := '0';
  
  SIGNAL prev_out_ready       : STD_LOGIC_VECTOR(0 TO g_output_rl);
  SIGNAL prev_out_data        : STD_LOGIC_VECTOR(c_data_w-1 DOWNTO 0);
  
  SIGNAL expected_out_data    : STD_LOGIC_VECTOR(c_data_w-1 DOWNTO 0);
    
  -- tb dut
  SIGNAL in_siso              : t_dp_siso;
  SIGNAL in_sosi              : t_dp_sosi;
  SIGNAL in_data              : STD_LOGIC_VECTOR(c_data_w-1 DOWNTO 0);
  SIGNAL in_sync              : STD_LOGIC;
  SIGNAL in_val               : STD_LOGIC;
  SIGNAL in_sop               : STD_LOGIC;
  SIGNAL in_eop               : STD_LOGIC;
  
  SIGNAL fifo_usedw           : STD_LOGIC_VECTOR(ceil_log2(2+g_input_rl+g_fifo_size)-1 DOWNTO 0);  -- +2+g_input_rl to match dp_latency_fifo usedw width
  SIGNAL fifo_ful             : STD_LOGIC;
  SIGNAL fifo_emp             : STD_LOGIC;
  
  SIGNAL out_siso             : t_dp_siso := c_dp_siso_hold;  -- default xon = '1' is needed for proc_dp_gen_block_data()
  SIGNAL out_sosi             : t_dp_sosi;
  SIGNAL out_data             : STD_LOGIC_VECTOR(c_data_w-1 DOWNTO 0);
  SIGNAL out_sync             : STD_LOGIC;
  SIGNAL out_val              : STD_LOGIC;
  SIGNAL out_sop              : STD_LOGIC;
  SIGNAL out_eop              : STD_LOGIC;
  
  SIGNAL out_gap              : STD_LOGIC := '1';
  
BEGIN

  clk <= (NOT clk) OR tb_end AFTER clk_period/2;
  rst <= '1', '0' AFTER clk_period*7;
  
  random_0 <= func_common_random(random_0) WHEN rising_edge(clk);
  random_1 <= func_common_random(random_1) WHEN rising_edge(clk);
  
  proc_common_gen_duty_pulse(c_pulse_0_active, c_pulse_0_period,   '1', rst, clk, pulse_en, pulse_0);
  proc_common_gen_duty_pulse(c_pulse_1_active, c_pulse_1_period+1, '1', rst, clk, pulse_en, pulse_1);

  ------------------------------------------------------------------------------
  -- STREAM CONTROL
  ------------------------------------------------------------------------------
  
  in_en          <= '1'                     WHEN g_in_en=e_active      ELSE
                    random_0(random_0'HIGH) WHEN g_in_en=e_random      ELSE
                    pulse_0                 WHEN g_in_en=e_pulse;
                       
  out_siso.ready <= '1'                     WHEN g_out_ready=e_active  ELSE
                    random_1(random_1'HIGH) WHEN g_out_ready=e_random  ELSE
                    pulse_1                 WHEN g_out_ready=e_pulse;
  
  
  ------------------------------------------------------------------------------
  -- DATA GENERATION
  ------------------------------------------------------------------------------
  
  -- Generate data path input data
  p_stimuli : PROCESS
    VARIABLE v_sync      : STD_LOGIC := '0';
    VARIABLE v_bsn       : STD_LOGIC_VECTOR(c_dp_stream_bsn_w-1 DOWNTO 0) := (OTHERS=>'0');
    VARIABLE v_channel   : NATURAL := 0;
    VARIABLE v_err       : NATURAL := 0;
    VARIABLE v_symbol    : NATURAL := c_symbol_init;
  BEGIN
    in_sosi <= c_dp_sosi_rst;
    proc_common_wait_until_low(clk, rst);
    proc_common_wait_some_cycles(clk, 5);
    
    -- Begin of stimuli
    FOR R IN 0 TO c_nof_repeat-1 LOOP
      proc_dp_gen_block_data(g_input_rl, TRUE, c_data_w, c_symbol_w, v_symbol, 0, 0, g_nof_symbols_per_block, v_channel, v_err, v_sync, v_bsn, clk, in_en, in_siso, in_sosi);
      v_bsn     := INCR_UVEC(v_bsn, 1);
      v_symbol  := (v_symbol + g_nof_symbols_per_block) MOD c_symbol_mod;
      v_channel := v_channel + 1;
      v_err     := v_err + 1;
      --proc_common_wait_some_cycles(clk, 10);               -- create gap between frames
    END LOOP;
    
    -- End of stimuli
    expected_out_data    <= in_sosi.data(c_data_w-1 DOWNTO 0);
    
    proc_common_wait_some_cycles(clk, 50);  -- depends on stream control
    verify_done <= '1';
    proc_common_wait_some_cycles(clk, 1);
    verify_done <= '0';
    proc_common_wait_some_cycles(clk, 10);
    tb_end <= '1';
    WAIT;
  END PROCESS;

      
  ------------------------------------------------------------------------------
  -- DATA VERIFICATION
  ------------------------------------------------------------------------------
  
  verify_en <= '1' WHEN rising_edge(clk) AND out_sosi.sop='1';  -- verify enable after first output sop
  
  -- SOSI control
  proc_dp_verify_valid(g_output_rl, clk, verify_en, out_siso.ready, prev_out_ready, out_val);        -- Verify that the output valid fits with the output ready latency
  proc_dp_verify_gap_invalid(clk, out_val, out_sop, out_eop, out_gap);                        -- Verify that the output valid is low between blocks from eop to sop
    
  -- SOSI data
  -- . verify that the output is incrementing symbols, like the input stimuli
  proc_dp_verify_symbols(g_output_rl, c_data_w, c_symbol_w, clk, verify_en, out_siso.ready, out_sosi.valid, out_sosi.eop, out_data, out_sosi.empty, prev_out_data);
    
  -- . verify that the stimuli have been applied at all
  proc_dp_verify_value(e_equal, clk, verify_done, expected_out_data, out_data);
  
  
  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------

  dut : ENTITY work.dp_latency_fifo
  GENERIC MAP (
    g_bypass     => g_bypass,
    g_input_rl   => g_input_rl,       -- input ready latency
    g_output_rl  => g_output_rl,      -- output ready latency
    g_fifo_size  => g_fifo_size
  )
  PORT MAP (
    rst         => rst,
    clk         => clk,
    -- Monitor FIFO filling
    usedw       => fifo_usedw,
    wr_ful      => fifo_ful,
    rd_emp      => fifo_emp,
    -- ST sink
    snk_in      => in_sosi,
    snk_out     => in_siso,
    -- ST source
    src_out     => out_sosi,
    src_in      => out_siso
  );
  
  -- Map to slv to ease monitoring in wave window
  in_data     <= in_sosi.data(c_data_w-1 DOWNTO 0);
  in_val      <= in_sosi.valid;
  in_sop      <= in_sosi.sop;
  in_eop      <= in_sosi.eop;
  in_sync     <= in_sosi.sync;

  out_data    <= out_sosi.data(c_data_w-1 DOWNTO 0);
  out_val     <= out_sosi.valid;
  out_sop     <= out_sosi.sop;
  out_eop     <= out_sosi.eop;
  out_sync    <= out_sosi.sync;
    
END tb;
