-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose:
-- . Test bench for dp_repack
-- Description:
--
-- Usage:
-- > as 10
-- > run -all
--
-- Remarks:
-- . This tb was made by copying and modifying tb_dp_packet_merge and with some knowledge from tb_dp_packetizing.
-- 

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_lfsr_sequences_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE work.tb_dp_pkg.ALL;


ENTITY tb_dp_repack IS
  GENERIC (
    -- general
    g_flow_control_stimuli   : t_dp_flow_control_enum := e_pulse;   -- always active, random or pulse flow control
    g_flow_control_verify    : t_dp_flow_control_enum := e_active;  -- always active, random or pulse flow control
    -- specific
--    g_in_dat_w               : NATURAL := 8;
--    g_in_nof_words           : NATURAL := 9;
--    g_pack_dat_w             : NATURAL := 24;
--    g_pack_nof_words         : NATURAL := 3;
--    g_nof_repeat             : NATURAL := 24;
--    g_pkt_len                : NATURAL := 36;  -- must be a multiple of g_in_nof_words
--    g_pkt_gap                : NATURAL := 4    -- must be >= g_pack_nof_words
    g_in_dat_w               : NATURAL := 8;
    g_in_nof_words           : NATURAL := 4;
    g_pack_dat_w             : NATURAL := 16;
    g_pack_nof_words         : NATURAL := 2;
    g_nof_repeat             : NATURAL := 24;
    g_pkt_len                : NATURAL := 20;  -- must be a multiple of g_in_nof_words
    g_pkt_gap                : NATURAL := 4    -- must be >= g_pack_nof_words
  );
END tb_dp_repack;


ARCHITECTURE tb OF tb_dp_repack IS

  -- dp_stream_stimuli
  CONSTANT c_stimuli_pulse_active     : NATURAL := 1;
  CONSTANT c_stimuli_pulse_period     : NATURAL := 2;
  
  CONSTANT c_data_init                : NATURAL := 0;
  CONSTANT c_bsn_init                 : STD_LOGIC_VECTOR(c_dp_stream_bsn_w-1 DOWNTO 0) := X"0000000000000000";  -- X"0877665544332211"
  CONSTANT c_err_init                 : NATURAL := 247;
  CONSTANT c_channel_init             : NATURAL := 5;  -- fixed
  
  -- dp_stream_verify
  CONSTANT c_verify_pulse_active      : NATURAL := g_in_nof_words;
  CONSTANT c_verify_pulse_period      : NATURAL := largest(g_in_nof_words, g_pack_nof_words);
  
  CONSTANT c_data_max                 : UNSIGNED(g_in_dat_w-1 DOWNTO 0) := (OTHERS=>'1');
  CONSTANT c_dsp_max                  : UNSIGNED(g_in_dat_w-1 DOWNTO 0) := (OTHERS=>'1');
  
  --CONSTANT c_verify_snk_in_cnt_max    : t_dp_sosi_unsigned := c_dp_sosi_unsigned_rst;  -- default 0 is no wrap
  CONSTANT c_verify_snk_in_cnt_max    : t_dp_sosi_unsigned := TO_DP_SOSI_UNSIGNED('0', '0', '0', '0', c_data_max, c_dsp_max, c_dsp_max, c_unsigned_0, c_unsigned_0, c_unsigned_0, c_unsigned_0);
  CONSTANT c_verify_snk_in_cnt_gap    : t_dp_sosi_unsigned := c_dp_sosi_unsigned_ones; -- default only accept increment +1
  
  -- both
  CONSTANT c_sync_period              : NATURAL := 10;
  CONSTANT c_sync_offset              : NATURAL := 7;
  
  SIGNAL clk                        : STD_LOGIC := '1';
  SIGNAL rst                        : STD_LOGIC := '1';
  SIGNAL tb_end                     : STD_LOGIC := '0';

  SIGNAL stimuli_src_in             : t_dp_siso;
  SIGNAL stimuli_src_out            : t_dp_sosi;
  SIGNAL stimuli_src_out_data       : STD_LOGIC_VECTOR(g_in_dat_w-1 DOWNTO 0);
  
  SIGNAL verify_snk_in_enable       : t_dp_sosi_sl := c_dp_sosi_sl_rst;
  SIGNAL last_snk_in                : t_dp_sosi;
  SIGNAL last_snk_in_evt            : STD_LOGIC;
  SIGNAL verify_last_snk_in_evt     : t_dp_sosi_sl := c_dp_sosi_sl_rst;
  
  SIGNAL verify_snk_out             : t_dp_siso := c_dp_siso_rdy;
  SIGNAL verify_snk_in              : t_dp_sosi;
  SIGNAL verify_snk_in_data         : STD_LOGIC_VECTOR(g_in_dat_w-1 DOWNTO 0);
  
  -- specific  
  SIGNAL pack_src_in                : t_dp_siso;
  SIGNAL pack_src_out               : t_dp_sosi;
  
  SIGNAL unpack_src_in              : t_dp_siso;
  SIGNAL unpack_src_out             : t_dp_sosi;
  
BEGIN

  clk <= (NOT clk) OR tb_end AFTER clk_period/2;
  rst <= '1', '0' AFTER clk_period*7;
  
  ------------------------------------------------------------------------------
  -- DATA GENERATION
  ------------------------------------------------------------------------------
  
  u_dp_stream_stimuli : ENTITY work.dp_stream_stimuli
  GENERIC MAP (
    g_instance_nr    => 0,                        -- only one stream so choose index 0
    -- flow control
    g_random_w       => 15,                       -- use different random width for stimuli and for verify to have different random sequences
    g_pulse_active   => c_stimuli_pulse_active,
    g_pulse_period   => c_stimuli_pulse_period,
    g_flow_control   => g_flow_control_stimuli,   -- always active, random or pulse flow control
    -- initializations
    g_sync_period    => c_sync_period,
    g_sync_offset    => c_sync_offset,
    g_data_init      => c_data_init,
    g_bsn_init       => c_bsn_init,
    g_err_init       => c_err_init,
    g_channel_init   => c_channel_init,
    -- specific
    g_in_dat_w       => g_in_dat_w,
    g_nof_repeat     => g_nof_repeat,  
    g_pkt_len        => g_pkt_len,
    g_pkt_gap        => g_pkt_gap     
  )
  PORT MAP (
    rst                 => rst,
    clk                 => clk,
  
    -- Generate stimuli
    src_in              => stimuli_src_in,
    src_out             => stimuli_src_out,

    -- End of stimuli
    last_snk_in         => last_snk_in,      -- expected verify_snk_in after end of stimuli 
    last_snk_in_evt     => last_snk_in_evt,  -- trigger verify to verify the last_snk_in 
    tb_end              => tb_end            -- signal end of tb as far as this dp_stream_stimuli is concerned
  );
  
  
  ------------------------------------------------------------------------------
  -- DATA VERIFICATION
  ------------------------------------------------------------------------------
  
  -- Select fields that need to be verified
  -- . during the test
  verify_snk_in_enable.sync    <= '0';
  verify_snk_in_enable.bsn     <= '0';
  verify_snk_in_enable.data    <= '1';
  verify_snk_in_enable.re      <= '0';
  verify_snk_in_enable.im      <= '0';
  verify_snk_in_enable.valid   <= '1';
  verify_snk_in_enable.sop     <= '1';
  verify_snk_in_enable.eop     <= '1';
  verify_snk_in_enable.empty   <= '0';
  verify_snk_in_enable.channel <= '0';
  verify_snk_in_enable.err     <= '0';
  
  -- . after the test
  verify_last_snk_in_evt.sync    <= '0';
  verify_last_snk_in_evt.bsn     <= '0';
  verify_last_snk_in_evt.data    <= last_snk_in_evt;
  verify_last_snk_in_evt.re      <= '0';
  verify_last_snk_in_evt.im      <= '0';
  verify_last_snk_in_evt.valid   <= last_snk_in_evt;
  verify_last_snk_in_evt.sop     <= last_snk_in_evt;
  verify_last_snk_in_evt.eop     <= last_snk_in_evt;
  verify_last_snk_in_evt.empty   <= '0';
  verify_last_snk_in_evt.channel <= '0';
  verify_last_snk_in_evt.err     <= '0';
  
  u_dp_stream_verify : ENTITY work.dp_stream_verify
  GENERIC MAP (
    g_instance_nr    => 0,                        -- only one stream so choose index 0
    -- flow control
    g_random_w       => 14,                       -- use different random width for stimuli and for verify to have different random sequences
    g_pulse_active   => c_verify_pulse_active,
    g_pulse_period   => c_verify_pulse_period,
    g_flow_control   => g_flow_control_verify,    -- always active, random or pulse flow control
    -- initializations
    g_sync_period    => c_sync_period,
    g_sync_offset    => c_sync_offset,
    g_snk_in_cnt_max => c_verify_snk_in_cnt_max,
    g_snk_in_cnt_gap => c_verify_snk_in_cnt_gap,
    -- specific
    g_in_dat_w       => g_in_dat_w,
    g_pkt_len        => g_pkt_len
  )
  PORT MAP (
    rst                        => rst,
    clk                        => clk,
  
    -- Verify data
    snk_out                    => verify_snk_out,
    snk_in                     => verify_snk_in,

    -- During stimuli
    verify_snk_in_enable       => verify_snk_in_enable,  -- enable verify to verify that the verify_snk_in fields are incrementing 
    
    -- End of stimuli
    expected_snk_in            => last_snk_in,            -- expected verify_snk_in after end of stimuli 
    verify_expected_snk_in_evt => verify_last_snk_in_evt  -- trigger verify to verify the last_snk_in 
  );  

  ------------------------------------------------------------------------------
  -- DUT Pack
  ------------------------------------------------------------------------------
  
  stimuli_src_in <= pack_src_in;
    
  u_pack : ENTITY work.dp_repack
  GENERIC MAP (
    g_in_dat_w       => g_in_dat_w,
    g_in_nof_words   => g_in_nof_words,
    g_out_dat_w      => g_pack_dat_w,
    g_out_nof_words  => g_pack_nof_words
  )
  PORT MAP (
    rst              => rst,
    clk              => clk,
    
    in_dat           => stimuli_src_out.data(g_in_dat_w-1 DOWNTO 0),
    in_val           => stimuli_src_out.valid,
    in_sof           => stimuli_src_out.sop,
    in_eof           => stimuli_src_out.eop,
    in_sync          => stimuli_src_out.sync,  -- DP style sync at sof
    
    out_dat          => pack_src_out.data(g_pack_dat_w-1 DOWNTO 0),
    out_val          => pack_src_out.valid,
    out_sof          => pack_src_out.sop,
    out_eof          => pack_src_out.eop,
    sof_sync         => pack_src_out.sync      -- DP style sync at sof, passes on in_sync
  );
  
  ------------------------------------------------------------------------------
  -- DUT Unpack
  ------------------------------------------------------------------------------
  
  pack_src_in <= unpack_src_in;
  
  u_unpack : ENTITY work.dp_repack
  GENERIC MAP (
    g_in_dat_w       => g_pack_dat_w,
    g_in_nof_words   => g_pack_nof_words,
    g_out_dat_w      => g_in_dat_w,
    g_out_nof_words  => g_in_nof_words
  )
  PORT MAP (
    rst              => rst,
    clk              => clk,
    
    in_dat           => pack_src_out.data(g_pack_dat_w-1 DOWNTO 0),
    in_val           => pack_src_out.valid,
    in_sof           => pack_src_out.sop,
    in_eof           => pack_src_out.eop,
    in_sync          => pack_src_out.sync,    -- DP style sync at sof
    
    out_dat          => unpack_src_out.data(g_in_dat_w-1 DOWNTO 0),
    out_val          => unpack_src_out.valid,
    out_sof          => unpack_src_out.sop,
    out_eof          => unpack_src_out.eop,
    sof_sync         => unpack_src_out.sync   -- DP style sync at sof, passes on in_sync
  );
  
  unpack_src_in <= verify_snk_out;
  
  p_verify_snk_in : PROCESS(unpack_src_out)
  BEGIN
    verify_snk_in       <= c_dp_sosi_rst;
    verify_snk_in.data  <= RESIZE_DP_DATA(unpack_src_out.data(g_in_dat_w-1 DOWNTO 0));
    verify_snk_in.valid <= unpack_src_out.valid;
    verify_snk_in.sop   <= unpack_src_out.sop;
    verify_snk_in.eop   <= unpack_src_out.eop;
    verify_snk_in.sync  <= unpack_src_out.sync;
  END PROCESS;
  
  ------------------------------------------------------------------------------
  -- Auxiliary
  ------------------------------------------------------------------------------
  
  -- Map to slv to ease monitoring in wave window
  stimuli_src_out_data <= stimuli_src_out.data(g_in_dat_w-1 DOWNTO 0);
  verify_snk_in_data   <= verify_snk_in.data(g_in_dat_w-1 DOWNTO 0);

END tb;
