--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Purpose:
-- . Provide an array of complex adders with streaming I/O
-- Description:
-- . Adds snk_in_2arr_2[i][0] to snk_in_2arr_2[i][1] yielding 
--   src_out_arr[i] for i in 0..g_nof_inputs-1.
-- . Valid is taken from snk_in_2arr_2(i)(0) only.

ENTITY corr_adder IS
  GENERIC (
    g_nof_inputs  : NATURAL;
    g_data_w      : NATURAL
   ); 
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_2arr_2  : IN  t_dp_sosi_2arr_2(g_nof_inputs-1 DOWNTO 0);

    src_out_arr   : OUT t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0)
  );
END corr_adder;

ARCHITECTURE str OF corr_adder IS

  CONSTANT c_input_pipeline_depth  : NATURAL := 1;
  CONSTANT c_output_pipeline_depth : NATURAL := 1;
  CONSTANT c_pipeline_depth        : NATURAL := c_input_pipeline_depth + c_output_pipeline_depth;

  -- Signal that carries the full adder output width
  SIGNAL common_complex_add_sub_src_out_arr : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  -- Same as above but minus the extra bit added by the adder.
  SIGNAL ranged_common_complex_add_sub_src_out_arr : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

BEGIN

  -----------------------------------------------------------------------------
  -- Delay valid signal
  -----------------------------------------------------------------------------
  gen_common_pipeline_sl : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    u_common_pipeline_sl : ENTITY common_lib.common_pipeline_sl
    GENERIC MAP (
      g_pipeline  => c_pipeline_depth
    )
    PORT MAP (
      rst     => rst,
      clk     => clk,
      in_dat  => snk_in_2arr_2(i)(0).valid,
      out_dat => src_out_arr(i).valid
    );
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Complex adder
  -----------------------------------------------------------------------------
  gen_common_complex_add_sub : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    u_common_complex_add_sub : ENTITY common_lib.common_complex_add_sub
    GENERIC MAP (
      g_direction       => "ADD",
      g_representation  => "SIGNED",
      g_pipeline_input  => c_input_pipeline_depth,
      g_pipeline_output => c_output_pipeline_depth,
      g_in_dat_w        => g_data_w,
      g_out_dat_w       => g_data_w+1
    )
    PORT MAP (
      clk        => clk,
      in_ar      => snk_in_2arr_2(i)(0).re(g_data_w-1 DOWNTO 0),
      in_ai      => snk_in_2arr_2(i)(0).im(g_data_w-1 DOWNTO 0),
      in_br      => snk_in_2arr_2(i)(1).re(g_data_w-1 DOWNTO 0),
      in_bi      => snk_in_2arr_2(i)(1).im(g_data_w-1 DOWNTO 0),
      out_re     => common_complex_add_sub_src_out_arr(i).re(g_data_w+1-1 DOWNTO 0),
      out_im     => common_complex_add_sub_src_out_arr(i).im(g_data_w+1-1 DOWNTO 0)
    );     
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Extend the sign bit
  -----------------------------------------------------------------------------
  gen_sign_extend : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    src_out_arr(i).re <= RESIZE_DP_DSP_DATA(common_complex_add_sub_src_out_arr(i).re(g_data_w-1 DOWNTO 0));
    src_out_arr(i).im <= RESIZE_DP_DSP_DATA(common_complex_add_sub_src_out_arr(i).im(g_data_w-1 DOWNTO 0));
  END GENERATE;

END str;
