--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Purpose:
-- . Outputs all unique pair permutations of the words at the inputs.
-- Description:
-- . c_nof_permutations = g_nof_inputs*(g_nof_inputs+1)/2
-- . Example permutations of [0,1,2]: [[0,0],[0,1],[0,2],[1,1],[1,2],[2,2]]
--   . Note the order of the permutations; low pairs to high pairs
-- . All inputs to permutate (snk_in_arr) must be valid on one and the same clock cycle.
-- . Process p_permu_wires is wires only and provides all permutations
--   instantaniously, in parallel:
--   . Low pair  [snk_in_arr[0],             snk_in_arr[0]             ] on permu_out_2arr_2[0]
--   . ..
--   . ..
--   . High pair [snk_in_arr[g_nof_inputs-1],snk_in_arr[g_nof_inputs-1]] on permu_out_2arr_2[g_nof_inputs-1]
-- . Both the input and the output of process p_permu_wires are registered by default.

ENTITY corr_permutator IS
  GENERIC (
    g_nof_inputs      : NATURAL := 8;
    g_register_input  : BOOLEAN := TRUE;
    g_register_output : BOOLEAN := TRUE
  ); 
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

    src_out_2arr_2 : OUT t_dp_sosi_2arr_2(g_nof_inputs*(g_nof_inputs+1)/2-1 DOWNTO 0) -- Array of pairs
  );
END corr_permutator;

ARCHITECTURE rtl OF corr_permutator IS

  CONSTANT c_nof_permutations : NATURAL := g_nof_inputs*(g_nof_inputs+1)/2;

  SIGNAL reg_snk_in_arr       : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  SIGNAL nxt_reg_snk_in_arr   : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

  TYPE t_permu_in_2arr IS ARRAY(g_nof_inputs-1 DOWNTO 0) OF t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  SIGNAL permu_in_2arr        : t_permu_in_2arr;
  SIGNAL permu_out_2arr_2     : t_dp_sosi_2arr_2(c_nof_permutations-1 DOWNTO 0);
  SIGNAL nxt_src_out_2arr_2   : t_dp_sosi_2arr_2(c_nof_permutations-1 DOWNTO 0);

BEGIN

  -----------------------------------------------------------------------------
  -- Input wiring: copy the inputs g_nof_inputs times
  -- . Direct wiring if we can output all permutations in parallel
  -- . Wire via input buffer if we need to serialize the output permutations
  -----------------------------------------------------------------------------
  gen_input: FOR i IN 0 TO g_nof_inputs-1 GENERATE
    gen_input: FOR j IN 0 TO g_nof_inputs-1 GENERATE
        permu_in_2arr(i)(j) <= reg_snk_in_arr(i);
    END GENERATE;
  END GENERATE;

  gen_input_register: IF g_register_input=TRUE GENERATE
    nxt_reg_snk_in_arr <= snk_in_arr;

    p_clk: PROCESS(clk, rst)
    BEGIN
      IF rst='1' THEN
        reg_snk_in_arr  <= (OTHERS=>c_dp_sosi_rst);
      ELSIF rising_edge(clk) THEN
        reg_snk_in_arr  <= nxt_reg_snk_in_arr;
      END IF;
    END PROCESS;
  END GENERATE;

  no_input_register: IF g_register_input=FALSE GENERATE
    reg_snk_in_arr <= snk_in_arr;
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Permutation wiring, 4-input example:
  -- ==================================== 
  --         0    1    2    3    4    5    6    7    8    9 <-- permu_out_2arr_2[c_nof_permutations][2]
  --  _      |    |    |    |    |    |    |    |    |    |
  -- |0|0-->[0]   |    |    |    |    |    |    |    |    |
  -- | |1------->[0]   |    |    |    |    |    |    |    |
  -- | |2---------|-->[0]   |    |    |    |    |    |    |
  -- |_|3---------|----|-->[0]   |    |    |    |    |    |
  -- |1|0------->[1]   |    |    |    |    |    |    |    |
  -- | |1--------------|----|-->[1]   |    |    |    |    |
  -- | |2--------------|----|------->[1]   |    |    |    |
  -- |_|3--------------|----|---------|-->[1]   |    |    |
  -- |2|0------------>[2]   |         |    |    |    |    |
  -- | |1-------------------|------->[2]   |    |    |    |
  -- | |2-------------------|--------------|-->[2]   |    |
  -- |_|3-------------------|--------------|------->[2]   |
  -- |3|0----------------->[3]             |         |    |
  -- | |1-------------------------------->[3]        |    |
  -- | |2------------------------------------------>[3]   |
  -- |_|3----------------------------------------------->[3]
  --  ^                                                     
  --  |
  --  permu_in_2arr[g_nof_inputs][g_nof_inputs]
  --
  -- Observe the following:
  -- . Inverse indexing: E.g. permutation[5], pair index 1 gets assigned 
  --   indices opposite to those of pair index 0:
  --   . permu_out_2arr_2(5)(0) <= permu_in_arr(1)(2);
  --     permu_out_2arr_2(5)(1) <= permu_in_arr(2)(1);
  -- . The number of inputs that carry duplicates increments by one 
  --   for each input increment (only pair index 0 shown):
  --   . permu_out_2arr_2(0)(0) <= permu_in_arr(0)(0); --permu_in_arr(0): 0 duplicates
  --     permu_out_2arr_2(1)(0) <= permu_in_arr(0)(1);  
  --     permu_out_2arr_2(2)(0) <= permu_in_arr(0)(2);
  --     permu_out_2arr_2(3)(0) <= permu_in_arr(0)(3);
  --     -- Duplicate              permu_in_arr(1)(0); --permu_in_arr(1): 1 duplicate
  --     permu_out_2arr_2(4)(0) <= permu_in_arr(1)(1);
  --     permu_out_2arr_2(5)(0) <= permu_in_arr(1)(2);
  --     permu_out_2arr_2(6)(0) <= permu_in_arr(1)(3);
  --     -- Duplicate              permu_in_arr(2)(0); --permu_in_arr(2): 2 duplicates
  --     -- Duplicate              permu_in_arr(2)(1);
  --     permu_out_2arr_2(7)(0) <= permu_in_arr(2)(2);
  --     permu_out_2arr_2(8)(0) <= permu_in_arr(2)(3);
  --     -- Duplicate              permu_in_arr(3)(0); --permu_in_arr(3): 3 duplicates
  --     -- Duplicate              permu_in_arr(3)(1);
  --     -- Duplicate              permu_in_arr(3)(2);
  --     permu_out_2arr_2(9)(0) <= permu_in_arr(3)(3);
  -----------------------------------------------------------------------------
  p_permu_wires: PROCESS(permu_in_2arr)
    VARIABLE v_out_index  : NATURAL;
    VARIABLE v_duplicates : NATURAL;
  BEGIN
    v_out_index  := 0;
    v_duplicates := 0;
    FOR i IN 0 TO g_nof_inputs-1 LOOP
      v_duplicates := i;
      FOR j IN 0 TO g_nof_inputs-1 LOOP
        -- Wire up our permutation pair by using inverse indices (j,i) and (i,j) 
        permu_out_2arr_2(v_out_index)(0) <= permu_in_2arr(i)(j);
        permu_out_2arr_2(v_out_index)(1) <= permu_in_2arr(j)(i);
        IF v_duplicates = 0 THEN
          -- We assigned a unique pair. Keep incrementing the index.
          v_out_index := v_out_index+1;
        ELSE
          -- We just wired a duplicate permutation pair. Don't increment 
          -- v_out_index so that permutation pair will get re-assigned in the
          -- next loop iteration.
          v_duplicates := v_duplicates-1;
        END IF;
      END LOOP;
    END LOOP;
  END PROCESS;

  -----------------------------------------------------------------------------
  -- Output wiring
  -----------------------------------------------------------------------------
  gen_output_register: IF g_register_output=TRUE GENERATE
    nxt_src_out_2arr_2 <= permu_out_2arr_2;

    p_clk: PROCESS(clk, rst)
    BEGIN
      IF rst='1' THEN
        src_out_2arr_2  <= (OTHERS=>(OTHERS=>c_dp_sosi_rst));
      ELSIF rising_edge(clk) THEN
        src_out_2arr_2  <= nxt_src_out_2arr_2;
      END IF;
    END PROCESS;
  END GENERATE;

  no_output_register: IF g_register_output=FALSE GENERATE  
    src_out_2arr_2 <= permu_out_2arr_2;
  END GENERATE;

END rtl;
