-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.eth_pkg.ALL;

-- Purpose:
--   Provide the total Ethernet header information for the user to decide upon.
-- Description:
--   Extract the first c_network_total_header_32b_nof_words=11 words from an
--   Ethernet packet and make them available via as a word array and as records
--   fields. The component has three output formats:
--   1) hdr_words_arr, the 11 word header store
--   2) hdr_data, the currently active header word
--   3) hdr_fields, combinatorial mapping of the 11 word header store to 
--      various Ethernet packet header fields.
--   Which format to use depends on the application. The logic for words or
--   header fields that are not used, will get optimized away during synthesis.
--   The snk_in word count is also output for the header words. The first valid
--   word marked by the snk_in.sop has count 0, the last header word has count
--   10 and the payload words have count 11.
-- Remarks:
-- . This component acts as a stream monitor, therefore it does not output a
--   snk_out.ready.
-- . The word indices from eth_pkg together with hdr_words_arr_val[] can be
--   used to determine which word of the Ethernet header words is valid.
-- . The word indices from eth_pkg together with hdr_fields_val[] can be used
--   to determine which fields of the Ethernet header records are valid. The
--   difference with hdr_words_arr_val is that hdr_words_arr_val is a single
--   cycle pulse, whereas hdr_fields_val keeps its state until the next sop.
--

ENTITY eth_hdr_store IS
  PORT (
    rst               : IN  STD_LOGIC;
    clk               : IN  STD_LOGIC;
    
    -- Streaming Sink
    snk_in            : IN  t_dp_sosi;
    snk_in_word_cnt   : OUT NATURAL RANGE 0 TO c_network_total_header_32b_nof_words;
    
    -- Total header
    hdr_words_arr     : OUT t_network_total_header_32b_arr;
    hdr_words_arr_val : OUT STD_LOGIC_VECTOR(0 TO c_network_total_header_32b_nof_words-1);
    -- Combinatorial map of the 11 header words on to the Ethernet header records
    hdr_fields        : OUT t_network_total_header;
    hdr_fields_val    : OUT STD_LOGIC_VECTOR(0 TO c_network_total_header_32b_nof_words-1);
    -- Support also outputting only the currently valid header data
    hdr_data          : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    hdr_data_val      : OUT STD_LOGIC
  );
END eth_hdr_store;


ARCHITECTURE rtl OF eth_hdr_store IS

  SIGNAL word_cnt              : NATURAL RANGE 0 TO c_network_total_header_32b_nof_words;
  SIGNAL nxt_word_cnt          : NATURAL;
  
  SIGNAL i_hdr_words_arr       : t_network_total_header_32b_arr := (OTHERS=>(OTHERS=>'0'));  -- init to avoid numeric_std warning, no need to rst
  SIGNAL nxt_hdr_words_arr     : t_network_total_header_32b_arr;
  SIGNAL nxt_hdr_words_arr_val : STD_LOGIC_VECTOR(hdr_fields_val'RANGE);
  SIGNAL i_hdr_fields_val      : STD_LOGIC_VECTOR(hdr_fields_val'RANGE);
  SIGNAL nxt_hdr_fields_val    : STD_LOGIC_VECTOR(hdr_fields_val'RANGE);
  SIGNAL i_hdr_data            : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0) := (OTHERS=>'0');  -- init to avoid numeric_std warning, no need to rst
  SIGNAL nxt_hdr_data          : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
  SIGNAL nxt_hdr_data_val      : STD_LOGIC;
  
BEGIN

  snk_in_word_cnt <= word_cnt;
  
  hdr_words_arr  <= i_hdr_words_arr;
  hdr_fields_val <= i_hdr_fields_val;
  hdr_data       <= i_hdr_data;
  
  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      -- input
      -- internal
      word_cnt          <= 0;
      -- output
      hdr_words_arr_val <= (OTHERS=>'0');
      i_hdr_fields_val  <= (OTHERS=>'0');
      hdr_data_val      <= '0';
    ELSIF rising_edge(clk) THEN
      -- input
      -- internal
      word_cnt          <= nxt_word_cnt;
      -- output
      i_hdr_words_arr   <= nxt_hdr_words_arr;      -- stored array of 11 words
      hdr_words_arr_val <= nxt_hdr_words_arr_val;
      i_hdr_fields_val  <= nxt_hdr_fields_val;
      i_hdr_data        <= nxt_hdr_data;       -- current word
      hdr_data_val      <= nxt_hdr_data_val;
    END IF;
  END PROCESS;
  
  -- Extract the 11 words from the Ethernet header
  p_words : PROCESS(word_cnt, i_hdr_words_arr, i_hdr_fields_val, i_hdr_data, snk_in)
  BEGIN
    nxt_word_cnt          <= word_cnt;
    nxt_hdr_words_arr     <= i_hdr_words_arr;
    nxt_hdr_words_arr_val <= (OTHERS=>'0');
    nxt_hdr_fields_val    <= i_hdr_fields_val;
    nxt_hdr_data          <= i_hdr_data;
    nxt_hdr_data_val      <= '0';
    IF snk_in.sop='1' THEN
      nxt_hdr_fields_val <= (OTHERS=>'0');
    END IF;
    IF snk_in.valid='1' AND word_cnt < c_network_total_header_32b_nof_words THEN
      -- new Ethernet header is arriving
      nxt_word_cnt                     <= word_cnt + 1;
      nxt_hdr_words_arr(     word_cnt) <= snk_in.data(c_eth_data_w-1 DOWNTO 0);
      nxt_hdr_words_arr_val( word_cnt) <= '1';
      nxt_hdr_fields_val(word_cnt)     <= '1';
      nxt_hdr_data                     <= snk_in.data(c_eth_data_w-1 DOWNTO 0);
      nxt_hdr_data_val                 <= '1';
    END IF;
    IF snk_in.eop='1' THEN
      nxt_word_cnt <= 0;
    END IF;
  END PROCESS;
  
  -- Combinatorial map of the total header on to the Ethernet header records
  hdr_fields.eth  <= func_network_total_header_extract_eth( i_hdr_words_arr);
  hdr_fields.arp  <= func_network_total_header_extract_arp( i_hdr_words_arr);
  hdr_fields.ip   <= func_network_total_header_extract_ip(  i_hdr_words_arr);
  hdr_fields.icmp <= func_network_total_header_extract_icmp(i_hdr_words_arr);
  hdr_fields.udp  <= func_network_total_header_extract_udp( i_hdr_words_arr);
  
END rtl;
