-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, tech_tse_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_tse_lib.tech_tse_pkg.ALL;


PACKAGE eth_pkg IS

  CONSTANT c_eth_data_w                : NATURAL := c_tech_tse_data_w;   -- = c_word_w
  CONSTANT c_eth_empty_w               : NATURAL := c_tech_tse_empty_w;  -- = ceil_log2(c_word_sz) = 2;
  CONSTANT c_eth_error_w               : NATURAL := c_tech_tse_error_w;  -- = 6, but no error field, pass error info on via checksum or CRC /= 0 in packet word
  
  CONSTANT c_eth_rx_ready_latency      : NATURAL := c_tech_tse_rx_ready_latency;  -- = 2 = default when FIFO is used
  CONSTANT c_eth_tx_ready_latency      : NATURAL := c_tech_tse_tx_ready_latency;  -- = 0, c_tech_tse_tx_ready_latency + 3 = TX_ALMOST_FULL
  CONSTANT c_eth_ready_latency         : NATURAL := 1;                       -- = 1, fixed ETH module internal RL

  -- Maximum feasible frame size
  CONSTANT c_eth_max_frame_sz          : NATURAL := 1024*9;  -- 9 kByte fits 9020 = word align (2) + eth header (14) + maximum jumbo payload (9000) + CRC (4)
  CONSTANT c_eth_max_frame_nof_words   : NATURAL := c_eth_max_frame_sz/c_word_sz;
  CONSTANT c_eth_max_frame_nof_words_w : NATURAL := ceil_log2(c_eth_max_frame_nof_words);  -- = 12 bit
  
  -- Actual frame space in RAM <= c_eth_max_frame_sz, default >= 1520, <= 9020 for jumbo
  --CONSTANT c_eth_frame_sz              : NATURAL := 1024*3/2;  -- Use 1536 = 3/2 M9K, to benefit from Rx and Tx in single buffer of 2*1.5=3 M9K, but does
                                                                 -- yield simulation warning: Address pointed at port A is out of bound!
  --CONSTANT c_eth_frame_sz              : NATURAL := 1024*8;  -- Use 8192 = 8 M9K for jumbo frame support with payload size <= 8172
  --CONSTANT c_eth_frame_sz              : NATURAL := 1024*9;  -- Use 9216 = 9 M9K for jumbo frame support with payload size <= 9000
  CONSTANT c_eth_frame_sz              : NATURAL := 1024*2;    -- Use 2048 = 2 M9K to avoid simulation warning: Address pointed at port A is out of bound!
                                                               -- when the module is used in an Nios II SOPC system
                                                               -- The FIFOs seem to require nof words to be a power of 2, but that is taken care of locally if necessary
  CONSTANT c_eth_frame_nof_words       : NATURAL := c_eth_frame_sz/c_word_sz;
  CONSTANT c_eth_frame_nof_words_w     : NATURAL := ceil_log2(c_eth_frame_nof_words);      -- >= 9 bit, <= 12 bit
                                                         
  TYPE t_eth_data_arr IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(c_eth_data_w-1 DOWNTO 0);
  
  
  ------------------------------------------------------------------------------
  -- Definitions for eth header status
  ------------------------------------------------------------------------------
  
  TYPE t_eth_hdr_status IS RECORD
    is_arp            : STD_LOGIC;
    is_ip             : STD_LOGIC;
    is_icmp           : STD_LOGIC;
    is_udp            : STD_LOGIC;
    ip_checksum       : STD_LOGIC_VECTOR(c_halfword_w-1 DOWNTO 0);
    ip_checksum_val   : STD_LOGIC;
    ip_checksum_is_ok : STD_LOGIC;
    udp_port          : STD_LOGIC_VECTOR(c_network_udp_port_slv'RANGE);
    is_dhcp           : STD_LOGIC;
  END RECORD;
  
  CONSTANT c_eth_hdr_status_rst : t_eth_hdr_status := ('0', '0', '0', '0',
                                                       (OTHERS=>'0'), '0', '0',
                                                       (OTHERS=>'0'), '0');
  
  ------------------------------------------------------------------------------
  -- Definitions for eth demux udp
  ------------------------------------------------------------------------------
  
  CONSTANT c_eth_nof_udp_ports : NATURAL := 4;                               -- support 2, 4, 8, ... UDP off-load ports
  CONSTANT c_eth_channel_w     : NATURAL := ceil_log2(c_eth_nof_udp_ports + 1);         -- + 1 for all other packets that go to the default port
  CONSTANT c_eth_nof_channels  : NATURAL := 2**c_eth_channel_w;
    
  ------------------------------------------------------------------------------
  -- MM register map
  ------------------------------------------------------------------------------
  
  CONSTANT c_eth_reg_demux_nof_words    : NATURAL := c_eth_nof_udp_ports;
  CONSTANT c_eth_reg_config_nof_words   : NATURAL := 4;
  CONSTANT c_eth_reg_control_nof_words  : NATURAL := 1;
  CONSTANT c_eth_reg_frame_nof_words    : NATURAL := 1;
  CONSTANT c_eth_reg_status_nof_words   : NATURAL := 1;
  
  CONSTANT c_eth_reg_demux_wi           : NATURAL := 0;
  CONSTANT c_eth_reg_config_wi          : NATURAL := c_eth_reg_demux_wi   + c_eth_reg_demux_nof_words;
  CONSTANT c_eth_reg_control_wi         : NATURAL := c_eth_reg_config_wi  + c_eth_reg_config_nof_words;
  CONSTANT c_eth_reg_frame_wi           : NATURAL := c_eth_reg_control_wi + c_eth_reg_control_nof_words;
  CONSTANT c_eth_reg_status_wi          : NATURAL := c_eth_reg_frame_wi   + c_eth_reg_frame_nof_words;
  CONSTANT c_eth_reg_continue_wi        : NATURAL := c_eth_reg_status_wi  + c_eth_reg_status_nof_words;

  -- . write/read back registers
  TYPE t_eth_demux_ports_arr IS ARRAY(1 TO c_eth_nof_udp_ports) OF STD_LOGIC_VECTOR(c_network_udp_port_w-1 DOWNTO 0);
  TYPE t_eth_mm_reg_demux IS RECORD
    udp_ports_en : STD_LOGIC_VECTOR(1 TO c_eth_nof_udp_ports);  -- [16]
    udp_ports    : t_eth_demux_ports_arr;                       -- [15:0]
  END RECORD;

  TYPE t_eth_mm_reg_config IS RECORD
    udp_port             : STD_LOGIC_VECTOR(c_network_udp_port_w-1 DOWNTO 0);       -- [15:0]
    ip_address           : STD_LOGIC_VECTOR(c_network_ip_addr_w-1 DOWNTO 0);        -- [31:0]
    mac_address          : STD_LOGIC_VECTOR(c_network_eth_mac_addr_w-1 DOWNTO 0);   -- [15:0], [31:0]
  END RECORD;
  
  TYPE t_eth_mm_reg_control IS RECORD
    tx_nof_words      : STD_LOGIC_VECTOR(c_eth_max_frame_nof_words_w-1 DOWNTO 0);  -- 12 bit
    tx_empty          : STD_LOGIC_VECTOR(c_eth_empty_w-1 DOWNTO 0);                -- 2 bit
    tx_request        : STD_LOGIC;                                                 -- 1 bit
    tx_en             : STD_LOGIC;                                                 -- 1 bit
    rx_en             : STD_LOGIC;                                                 -- 1 bit
  END RECORD;
  
  TYPE t_eth_mm_reg_control_bi IS RECORD  -- bit indices
    tx_nof_words      : NATURAL;          -- [26:18]    
    tx_empty          : NATURAL;          -- [17:16]    
    tx_request        : NATURAL;          -- [2]        
    tx_en             : NATURAL;          -- [1]        
    rx_en             : NATURAL;          -- [0]        
  END RECORD;
  
  CONSTANT c_eth_mm_reg_control_bi  : t_eth_mm_reg_control_bi := (18, 16, 2, 1, 0);
  CONSTANT c_eth_mm_reg_control_rst : t_eth_mm_reg_control := ((OTHERS=>'0'), (OTHERS=>'0'), '0', '0', '0');

  -- . read only registers
  TYPE t_eth_mm_reg_frame IS RECORD
    is_dhcp           : STD_LOGIC;
    is_udp_ctrl_port  : STD_LOGIC;
    is_udp            : STD_LOGIC;
    is_icmp           : STD_LOGIC;
    ip_address_match  : STD_LOGIC;
    ip_checksum_is_ok : STD_LOGIC;
    is_ip             : STD_LOGIC;
    is_arp            : STD_LOGIC;
    mac_address_match : STD_LOGIC;
    eth_mac_error     : STD_LOGIC_VECTOR(c_eth_error_w-1 DOWNTO 0);
  END RECORD;
  
  TYPE t_eth_mm_reg_frame_bi IS RECORD  -- bit indices
    is_dhcp           : NATURAL;        -- [15]
    is_udp_ctrl_port  : NATURAL;        -- [14]
    is_udp            : NATURAL;        -- [13]
    is_icmp           : NATURAL;        -- [12]
    ip_address_match  : NATURAL;        -- [11]
    ip_checksum_is_ok : NATURAL;        -- [10]
    is_ip             : NATURAL;        -- [9]
    is_arp            : NATURAL;        -- [8]
    mac_address_match : NATURAL;        -- [7]
    eth_mac_error     : NATURAL;        -- [6] not used, [5:0] = TSE MAC error
  END RECORD;
  
  CONSTANT c_eth_mm_reg_frame_bi  : t_eth_mm_reg_frame_bi := (15, 14, 13, 12, 11, 10, 9, 8, 7, 0);
  CONSTANT c_eth_mm_reg_frame_rst : t_eth_mm_reg_frame    := ('0', '0', '0', '0', '0', '0', '0', '0', '0', (OTHERS=>'0'));
    
  TYPE t_eth_mm_reg_status IS RECORD
    rx_nof_words      : STD_LOGIC_VECTOR(c_eth_max_frame_nof_words_w-1 DOWNTO 0);  -- 12 bit
    rx_empty          : STD_LOGIC_VECTOR(c_eth_empty_w-1 DOWNTO 0);                -- 2 bit
    tx_avail          : STD_LOGIC;                                                 -- 1 bit
    tx_done           : STD_LOGIC;                                                 -- 1 bit
    rx_avail          : STD_LOGIC;                                                 -- 1 bit
  END RECORD;
  
  TYPE t_eth_mm_reg_status_bi IS RECORD  -- bit indices
    rx_nof_words      : NATURAL;         -- [26:18]
    rx_empty          : NATURAL;         -- [17:16]
    tx_avail          : NATURAL;         -- [2]
    tx_done           : NATURAL;         -- [1]
    rx_avail          : NATURAL;         -- [0]
  END RECORD;
  
  CONSTANT c_eth_mm_reg_status_bi  : t_eth_mm_reg_status_bi := (18, 16, 2, 1, 0);
  CONSTANT c_eth_mm_reg_status_rst : t_eth_mm_reg_status    := ((OTHERS=>'0'), (OTHERS=>'0'), '0', '0', '0');
  
  -- Register mapping functions
  FUNCTION func_eth_mm_reg_demux(  mm_reg : STD_LOGIC_VECTOR)     RETURN t_eth_mm_reg_demux;
  FUNCTION func_eth_mm_reg_demux(  mm_reg : t_eth_mm_reg_demux)   RETURN STD_LOGIC_VECTOR;
  FUNCTION func_eth_mm_reg_config( mm_reg : STD_LOGIC_VECTOR)     RETURN t_eth_mm_reg_config;
  FUNCTION func_eth_mm_reg_config( mm_reg : t_eth_mm_reg_config)  RETURN STD_LOGIC_VECTOR;
  FUNCTION func_eth_mm_reg_control(mm_reg : STD_LOGIC_VECTOR)     RETURN t_eth_mm_reg_control;
  FUNCTION func_eth_mm_reg_control(mm_reg : t_eth_mm_reg_control) RETURN STD_LOGIC_VECTOR;
  FUNCTION func_eth_mm_reg_frame(  mm_reg : STD_LOGIC_VECTOR)     RETURN t_eth_mm_reg_frame;
  FUNCTION func_eth_mm_reg_frame(  mm_reg : t_eth_mm_reg_frame)   RETURN STD_LOGIC_VECTOR;
  FUNCTION func_eth_mm_reg_status( mm_reg : STD_LOGIC_VECTOR)     RETURN t_eth_mm_reg_status;
  FUNCTION func_eth_mm_reg_status( mm_reg : t_eth_mm_reg_status)  RETURN STD_LOGIC_VECTOR;
  
  ------------------------------------------------------------------------------
  -- Definitions for eth_mm_registers
  ------------------------------------------------------------------------------
  
  CONSTANT c_eth_reg_nof_words    : NATURAL := c_eth_reg_demux_nof_words +
                                               c_eth_reg_config_nof_words +
                                               c_eth_reg_control_nof_words +
                                               c_eth_reg_frame_nof_words +
                                               c_eth_reg_status_nof_words;
  CONSTANT c_eth_reg_addr_w       : NATURAL := ceil_log2(c_eth_reg_nof_words + 1);  -- + 1 for c_eth_continue_wi
  
  ------------------------------------------------------------------------------
  -- Definitions for ETH Rx packet buffer and Tx packet buffer
  ------------------------------------------------------------------------------
  
  -- Use MM bus data width = c_word_w = 32
  CONSTANT c_eth_ram_rx_offset : NATURAL := 0;
  CONSTANT c_eth_ram_tx_offset : NATURAL := c_eth_frame_nof_words;
  CONSTANT c_eth_ram_nof_words : NATURAL := c_eth_frame_nof_words*2;
  CONSTANT c_eth_ram_addr_w    : NATURAL := ceil_log2(c_eth_ram_nof_words);
  
END eth_pkg;


PACKAGE BODY eth_pkg IS
  
  -- Register mapping functions
  FUNCTION func_eth_mm_reg_demux(mm_reg : STD_LOGIC_VECTOR) RETURN t_eth_mm_reg_demux IS
    VARIABLE v_reg : t_eth_mm_reg_demux;
  BEGIN
    -- Demux UDP MM registers
    FOR I IN 1 TO c_eth_nof_udp_ports LOOP
      v_reg.udp_ports_en(I) := mm_reg(c_network_udp_port_w   + (I-1)*c_word_w);                        -- [16]   = UDP port enable
      v_reg.udp_ports(I)    := mm_reg(c_network_udp_port_w-1 + (I-1)*c_word_w DOWNTO (I-1)*c_word_w);  -- [15:0] = UDP port number
    END LOOP;
    RETURN v_reg;
  END func_eth_mm_reg_demux;

  FUNCTION func_eth_mm_reg_demux(mm_reg : t_eth_mm_reg_demux) RETURN STD_LOGIC_VECTOR IS
    VARIABLE v_reg : STD_LOGIC_VECTOR(c_eth_reg_demux_nof_words*c_word_w-1 DOWNTO 0);
  BEGIN
    v_reg := (OTHERS=>'0');                                                                            -- rsvd
    FOR I IN 1 TO c_eth_nof_udp_ports LOOP
      v_reg(c_network_udp_port_w   + (I-1)*c_word_w)                       := mm_reg.udp_ports_en(I);  -- [16]   = UDP port enable
      v_reg(c_network_udp_port_w-1 + (I-1)*c_word_w DOWNTO (I-1)*c_word_w) := mm_reg.udp_ports(I);     -- [15:0] = UDP port number
    END LOOP;
    RETURN v_reg;
  END func_eth_mm_reg_demux;
  
  -- MM config register
  FUNCTION func_eth_mm_reg_config(mm_reg : STD_LOGIC_VECTOR) RETURN t_eth_mm_reg_config IS
    VARIABLE v_reg : t_eth_mm_reg_config;
  BEGIN
    v_reg.udp_port                        := mm_reg(c_network_udp_port_w+3*c_word_w-1 DOWNTO 3*c_word_w);  -- [15:0]  = control UDP port number
    v_reg.ip_address                      := mm_reg(                     3*c_word_w-1 DOWNTO 2*c_word_w);  -- [31:0]  = this node IP address
    v_reg.mac_address(47 DOWNTO 32)       := mm_reg(        c_halfword_w+  c_word_w-1 DOWNTO   c_word_w);  -- [47:32] = this node MAC address
    v_reg.mac_address(31 DOWNTO 0)        := mm_reg(               c_word_w-1 DOWNTO   0);         -- [31:0]  = this node MAC address
    RETURN v_reg;
  END func_eth_mm_reg_config;

  FUNCTION func_eth_mm_reg_config(mm_reg : t_eth_mm_reg_config) RETURN STD_LOGIC_VECTOR IS
    VARIABLE v_reg : STD_LOGIC_VECTOR(c_eth_reg_config_nof_words*c_word_w-1 DOWNTO 0);
  BEGIN
    v_reg                                                      := (OTHERS=>'0');                     -- rsvd
    v_reg(c_network_udp_port_w+3*c_word_w-1 DOWNTO 3*c_word_w) := mm_reg.udp_port;                   -- [15:0]  = control UDP port number
    v_reg(                     3*c_word_w-1 DOWNTO 2*c_word_w) := mm_reg.ip_address;                 -- [31:0]  = this node IP address
    v_reg(        c_halfword_w+  c_word_w-1 DOWNTO   c_word_w) := mm_reg.mac_address(47 DOWNTO 32);  -- [47:32] = this node MAC address
    v_reg(                       c_word_w-1 DOWNTO   0)        := mm_reg.mac_address(31 DOWNTO 0);   -- [31:0]  = this node MAC address
    RETURN v_reg;
  END func_eth_mm_reg_config;
    
  -- MM control register
  FUNCTION func_eth_mm_reg_control(mm_reg : STD_LOGIC_VECTOR) RETURN t_eth_mm_reg_control IS
    VARIABLE v_reg : t_eth_mm_reg_control;
  BEGIN
    v_reg.tx_nof_words := mm_reg(c_eth_max_frame_nof_words_w+c_eth_empty_w+16-1 DOWNTO c_eth_empty_w+16);  -- [29:18]
    v_reg.tx_empty     := mm_reg(                            c_eth_empty_w+16-1 DOWNTO               16);  -- [17:16]
    v_reg.tx_request   := mm_reg(                                                                     2);  -- [2]
    v_reg.tx_en        := mm_reg(                                                                     1);  -- [1]
    v_reg.rx_en        := mm_reg(                                                                     0);  -- [0]
    RETURN v_reg;
  END func_eth_mm_reg_control;

  FUNCTION func_eth_mm_reg_control(mm_reg : t_eth_mm_reg_control) RETURN STD_LOGIC_VECTOR IS
    VARIABLE v_reg : STD_LOGIC_VECTOR(c_eth_reg_control_nof_words*c_word_w-1 DOWNTO 0);
  BEGIN
    v_reg                                                                         := (OTHERS=>'0');        -- rsvd
    v_reg(c_eth_max_frame_nof_words_w+c_eth_empty_w+16-1 DOWNTO c_eth_empty_w+16) := mm_reg.tx_nof_words;  -- [29:18]
    v_reg(                            c_eth_empty_w+16-1 DOWNTO               16) := mm_reg.tx_empty;      -- [17:16]
    v_reg(                                                                     2) := mm_reg.tx_request;    -- [2]
    v_reg(                                                                     1) := mm_reg.tx_en;         -- [1]
    v_reg(                                                                     0) := mm_reg.rx_en;         -- [0]
    RETURN v_reg;
  END func_eth_mm_reg_control;
    
  -- MM frame register
  FUNCTION func_eth_mm_reg_frame(mm_reg : STD_LOGIC_VECTOR) RETURN t_eth_mm_reg_frame IS
    VARIABLE v_reg : t_eth_mm_reg_frame;
  BEGIN
    v_reg.is_dhcp           := mm_reg(              c_byte_w+7);  -- [15]
    v_reg.is_udp_ctrl_port  := mm_reg(              c_byte_w+6);  -- [14]
    v_reg.is_udp            := mm_reg(              c_byte_w+5);  -- [13]
    v_reg.is_icmp           := mm_reg(              c_byte_w+4);  -- [12]
    v_reg.ip_address_match  := mm_reg(              c_byte_w+3);  -- [11]
    v_reg.ip_checksum_is_ok := mm_reg(              c_byte_w+2);  -- [10]
    v_reg.is_ip             := mm_reg(              c_byte_w+1);  -- [9]
    v_reg.is_arp            := mm_reg(              c_byte_w+0);  -- [8]
    v_reg.mac_address_match := mm_reg(              c_byte_w-1);  -- [7]
    v_reg.eth_mac_error     := mm_reg(c_eth_error_w-1 DOWNTO 0);  -- [7] not used, [5:0] = TSE MAC error
    RETURN v_reg;
  END func_eth_mm_reg_frame;
  
  FUNCTION func_eth_mm_reg_frame(mm_reg : t_eth_mm_reg_frame) RETURN STD_LOGIC_VECTOR IS
    VARIABLE v_reg : STD_LOGIC_VECTOR(c_eth_reg_frame_nof_words*c_word_w-1 DOWNTO 0);
  BEGIN
    v_reg                           := (OTHERS=>'0');             -- rsvd
    v_reg(              c_byte_w+7) := mm_reg.is_dhcp;            -- [15]
    v_reg(              c_byte_w+6) := mm_reg.is_udp_ctrl_port;   -- [14]
    v_reg(              c_byte_w+5) := mm_reg.is_udp;             -- [13]
    v_reg(              c_byte_w+4) := mm_reg.is_icmp;            -- [12]
    v_reg(              c_byte_w+3) := mm_reg.ip_address_match;   -- [11]
    v_reg(              c_byte_w+2) := mm_reg.ip_checksum_is_ok;  -- [10]
    v_reg(              c_byte_w+1) := mm_reg.is_ip;              -- [9]
    v_reg(              c_byte_w+0) := mm_reg.is_arp;             -- [8]
    v_reg(              c_byte_w-1) := mm_reg.mac_address_match;  -- [7]
    v_reg(c_eth_error_w-1 DOWNTO 0) := mm_reg.eth_mac_error;      -- [6] not used, [5:0] = TSE MAC error
    RETURN v_reg;
  END func_eth_mm_reg_frame;
  
  -- MM status register
  FUNCTION func_eth_mm_reg_status(mm_reg : STD_LOGIC_VECTOR) RETURN t_eth_mm_reg_status IS
    VARIABLE v_reg : t_eth_mm_reg_status;
  BEGIN
    v_reg.rx_nof_words := mm_reg(c_eth_max_frame_nof_words_w+c_eth_empty_w+16-1 DOWNTO c_eth_empty_w+16);  -- [29:18]
    v_reg.rx_empty     := mm_reg(                            c_eth_empty_w+16-1 DOWNTO               16);  -- [17:16]
    v_reg.tx_avail     := mm_reg(                                                                     2);  -- [2]
    v_reg.tx_done      := mm_reg(                                                                     1);  -- [1]
    v_reg.rx_avail     := mm_reg(                                                                     0);  -- [0]
    RETURN v_reg;
  END func_eth_mm_reg_status;

  FUNCTION func_eth_mm_reg_status(mm_reg : t_eth_mm_reg_status) RETURN STD_LOGIC_VECTOR IS
    VARIABLE v_reg : STD_LOGIC_VECTOR(c_eth_reg_status_nof_words*c_word_w-1 DOWNTO 0);
  BEGIN
    v_reg                                                                         := (OTHERS=>'0');        -- rsvd
    v_reg(c_eth_max_frame_nof_words_w+c_eth_empty_w+16-1 DOWNTO c_eth_empty_w+16) := mm_reg.rx_nof_words;  -- [29:18]
    v_reg(                            c_eth_empty_w+16-1 DOWNTO               16) := mm_reg.rx_empty;      -- [17:16]
    v_reg(                                                                     2) := mm_reg.tx_avail;      -- [2]
    v_reg(                                                                     1) := mm_reg.tx_done;       -- [1]
    v_reg(                                                                     0) := mm_reg.rx_avail;      -- [0]
    RETURN v_reg;
  END func_eth_mm_reg_status;
  
END eth_pkg;

