--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, mdio_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE mdio_lib.mdio_pkg.ALL;
USE mdio_lib.mdio_vitesse_vsc8486_pkg.ALL;

ENTITY tr_xaui_mdio IS
  GENERIC (
    g_sim                   : BOOLEAN := FALSE;
    g_nof_xaui              : NATURAL := 1;     -- Up to 3 (hard XAUI only) supported
    g_mdio_epcs_dis         : BOOLEAN := FALSE -- TRUE disables EPCS on init; e.g. to target a 10GbE card in PC that does not support it 
  );
  PORT (   
    -- Transceiver PLL reference clock   
    tr_clk                  : IN  STD_LOGIC;
    tr_rst                  : IN  STD_LOGIC;

    -- MM clock for register of optional MDIO master
    mm_clk                  : IN  STD_LOGIC := '0';
    mm_rst                  : IN  STD_LOGIC := '0';

    -- MDIO master = mm slave
    mdio_mosi_arr           : IN  t_mem_mosi_arr(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>c_mem_mosi_rst);
    mdio_miso_arr           : OUT t_mem_miso_arr(g_nof_xaui-1 DOWNTO 0);

    -- MDIO External clock and serial data.
    mdio_rst                : OUT STD_LOGIC;
    mdio_mdc_arr            : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);       
    mdio_mdat_in_arr        : IN  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>'0');
    mdio_mdat_oen_arr       : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0)
  );
END tr_xaui_mdio;


ARCHITECTURE str OF tr_xaui_mdio IS  

  CONSTANT c_mdio_xaui_phy_mdc_period     : NATURAL := 256;  -- must be a power of 2
  CONSTANT c_mdio_xaui_phy_hold_time      : NATURAL := 10;
  CONSTANT c_mdio_xaui_phy_setup_time     : NATURAL := 2;

  CONSTANT c_mdio_xaui_phy                : t_c_mdio_phy := (c_mdio_xaui_phy_mdc_period, c_mdio_xaui_phy_hold_time, c_mdio_xaui_phy_setup_time);

  SIGNAL mdio_hdr                  : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mdio_tx_dat               : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mdio_rx_dat               : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mdio_en_evt               : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mdio_done                 : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mdio_done_ack_evt         : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL reg_mdio_hdr              : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL reg_mdio_tx_dat           : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL reg_mdio_rx_dat           : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL reg_mdio_en_evt           : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL reg_mdio_done             : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL reg_mdio_done_ack_evt     : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL ctlr_mdio_rst             : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_hdr             : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_tx_dat          : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_rx_dat          : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_en_evt          : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_done            : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_done_ack_evt    : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL ctlr_exec_complete        : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

BEGIN
 
  -- UniBoard FN resets 4 Vitesse chips using one output.
  mdio_rst <= ctlr_mdio_rst(0);

  gen_nof_xaui : FOR i IN g_nof_xaui-1 DOWNTO 0 GENERATE

    -- PHY core
    u_mdio_phy : ENTITY mdio_lib.mdio_phy
    GENERIC MAP(
      g_mdio_phy      => c_mdio_xaui_phy,
      g_add_mdc_cycle => TRUE
    )
    PORT MAP (
      gs_sim            => g_sim,
      
      rst               => tr_rst,
      clk               => tr_clk,
    
      mdio_en_evt       => mdio_en_evt(i),
      mdio_done         => mdio_done(i),
      mdio_done_ack_evt => mdio_done_ack_evt(i),
    
      hdr               => mdio_hdr(i),
      tx_dat            => mdio_tx_dat(i),
      rx_dat            => mdio_rx_dat(i),
    
      -- External clock and serial data
      mdc               => mdio_mdc_arr(i),
      mdat_in           => mdio_mdat_in_arr(i),
      mdat_oen          => mdio_mdat_oen_arr(i)
    );

    -- MM port
    u_mdio_phy_reg : ENTITY mdio_lib.mdio_phy_reg
    PORT MAP ( 
      mm_rst            => mm_rst,
      mm_clk            => mm_clk,
  
      mdio_rst          => tr_rst,
      mdio_clk          => tr_clk,
  
      sla_in            => mdio_mosi_arr(i),
      sla_out           => mdio_miso_arr(i),
  
      mdio_en_evt       => reg_mdio_en_evt(i),
      mdio_done_ack_evt => reg_mdio_done_ack_evt(i),
      mdio_done         => reg_mdio_done(i),
  
      mdio_hdr          => reg_mdio_hdr(i),
      mdio_tx_dat       => reg_mdio_tx_dat(i),
      mdio_rx_dat       => reg_mdio_rx_dat(i)
    );

    -- MDIO controller auto-executes MDIO sequence on startup
    gen_mdio_epcs : IF g_mdio_epcs_dis = FALSE GENERATE
      u_mdio_ctlr_epcs : ENTITY mdio_lib.mdio_ctlr
      GENERIC MAP (
         g_mdio_prtad           => c_mdio_vsc8486_prtad,
         g_mdio_cmd_arr         => c_mdio_vsc8486_init_cmd_arr,
         g_mdio_rst_level       => '0', 
         g_mdio_rst_cycles      => sel_a_b(g_sim, 10, 250000), 
         g_mdio_post_rst_cycles => sel_a_b(g_sim, 10, 250000)
        ) 
      PORT MAP (
        rst               => tr_rst,
        clk               => tr_clk,

        mdio_rst          => ctlr_mdio_rst(i),
        mdio_en_evt       => ctlr_mdio_en_evt(i),
        mdio_done         => ctlr_mdio_done(i),
    
        mdio_done_ack_evt => ctlr_mdio_done_ack_evt(i),
    
        hdr               => ctlr_mdio_hdr(i),
        tx_dat            => ctlr_mdio_tx_dat(i),

        exec_complete     => ctlr_exec_complete(i)
      );
    END GENERATE;

    -- MDIO controller auto-executes MDIO sequence on startup
    gen_mdio_no_epcs : IF g_mdio_epcs_dis = TRUE GENERATE
      u_mdio_ctlr_no_epcs : ENTITY mdio_lib.mdio_ctlr
      GENERIC MAP (
         g_mdio_prtad           => c_mdio_vsc8486_prtad,
         g_mdio_cmd_arr         => c_mdio_vsc8486_init_epcs_dis_cmd_arr,
         g_mdio_rst_level       => '0', 
         g_mdio_rst_cycles      => sel_a_b(g_sim, 10, 250000),
         g_mdio_post_rst_cycles => sel_a_b(g_sim, 10, 250000)
        ) 
      PORT MAP (
        rst               => tr_rst,
        clk               => tr_clk,

        mdio_rst          => ctlr_mdio_rst(i),      
        mdio_en_evt       => ctlr_mdio_en_evt(i),
        mdio_done         => ctlr_mdio_done(i),
    
        mdio_done_ack_evt => ctlr_mdio_done_ack_evt(i),
    
        hdr               => ctlr_mdio_hdr(i),
        tx_dat            => ctlr_mdio_tx_dat(i),

        exec_complete     => ctlr_exec_complete(i)
      );
    END GENERATE;

    -- Connect the mdio_ctlr to the mdio_phy initially, when it's done connect the MM controller to allow user control/monitoring.
    mdio_en_evt(i)       <= ctlr_mdio_en_evt(i)       WHEN ctlr_exec_complete(i)='0' ELSE reg_mdio_en_evt(i); 
    mdio_done_ack_evt(i) <= ctlr_mdio_done_ack_evt(i) WHEN ctlr_exec_complete(i)='0' ELSE reg_mdio_done_ack_evt(i);
    mdio_hdr(i)          <= ctlr_mdio_hdr(i)          WHEN ctlr_exec_complete(i)='0' ELSE reg_mdio_hdr(i);
    mdio_tx_dat(i)       <= ctlr_mdio_tx_dat(i)       WHEN ctlr_exec_complete(i)='0' ELSE reg_mdio_tx_dat(i);       
  
    ctlr_mdio_done(i)    <= mdio_done(i)   WHEN ctlr_exec_complete(i)='0' ELSE '0';
    reg_mdio_done(i)     <= mdio_done(i)   WHEN ctlr_exec_complete(i)='1' ELSE '0';
    reg_mdio_rx_dat(i)   <= mdio_rx_dat(i) WHEN ctlr_exec_complete(i)='1' ELSE (OTHERS=>'0');

  END GENERATE;
      
END str;

