--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Declare IP libraries to ensure default binding in simulation. The IP library clause is ignored by synthesis.
LIBRARY ip_stratixiv_transceiver_lib;

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE work.tech_transceiver_component_pkg.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

ENTITY tech_transceiver_gx_stratix_iv IS 
  GENERIC(
    g_data_w                : NATURAL := 32; -- Supported: 32, 16.
    g_nof_gx                : NATURAL;
    g_mbps                  : NATURAL;  -- Supported: 6250, 5000, 3125, 2500 when g_data_w=32, else 2500 when g_data_w=16
    g_sim                   : BOOLEAN;
    g_tx                    : BOOLEAN;
    g_rx                    : BOOLEAN
  );      
  PORT(
    cal_rec_clk             : IN  STD_LOGIC;
    tr_clk                  : IN  STD_LOGIC;

    rx_clk                  : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    rx_rst                  : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    rx_sosi_arr             : OUT t_dp_sosi_arr(g_nof_gx-1 DOWNTO 0);
    rx_siso_arr             : IN  t_dp_siso_arr(g_nof_gx-1 DOWNTO 0);

    tx_clk                  : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    tx_rst                  : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    tx_sosi_arr             : IN  t_dp_sosi_arr(g_nof_gx-1 DOWNTO 0);
    tx_siso_arr             : OUT t_dp_siso_arr(g_nof_gx-1 DOWNTO 0);
   
    --Serial I/O
    rx_datain               : IN  STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    tx_dataout              : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);   

    -- Ctrl&stat
    tx_state                : OUT STD_LOGIC_VECTOR(2*g_nof_gx-1 DOWNTO 0);
    tx_align_en             : IN  STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);

    rx_state                : OUT STD_LOGIC_VECTOR(2*g_nof_gx-1 DOWNTO 0);
    rx_align_en             : IN  STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0)
  );

END tech_transceiver_gx_stratix_iv;

ARCHITECTURE str OF tech_transceiver_gx_stratix_iv IS

  CONSTANT c_align_pattern_32      : STD_LOGIC_VECTOR := x"0000BC1C";  --Change in megafunction first. BC is std. comma symbol (K.28.5).
  CONSTANT c_inval_pattern_32      : STD_LOGIC_VECTOR := x"3C3C3C3C";  --comma symbol K.28.1 (decoded as 3C) is used as invalid data indicator.

  CONSTANT c_align_pattern_16      : STD_LOGIC_VECTOR := x"BCBC";      --Change in megafunction first. BC is std. comma symbol (K.28.5).
  CONSTANT c_inval_pattern_16      : STD_LOGIC_VECTOR := x"BCBC";

  CONSTANT c_align_pattern         : STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0) := sel_a_b(g_data_w=16, c_align_pattern_16, c_align_pattern_32);
  CONSTANT c_inval_pattern         : STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0) := sel_a_b(g_data_w=16, c_inval_pattern_16, c_inval_pattern_32);

  CONSTANT c_reconf_togxb_bus_w    : NATURAL := 4;
  CONSTANT c_reconf_fromgxb_bus_w  : NATURAL := 17;

  CONSTANT c_nof_ctrl_bytes        : NATURAL := g_data_w / c_byte_w;

  TYPE t_data_arr IS ARRAY (NATURAL RANGE <>) OF STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0);
  TYPE t_ctrl_arr IS ARRAY (NATURAL RANGE <>) OF STD_LOGIC_VECTOR(c_nof_ctrl_bytes-1 DOWNTO 0);

  SIGNAL reconfig_togxb            : STD_LOGIC_VECTOR(c_reconf_togxb_bus_w-1 DOWNTO 0);
  SIGNAL reconfig_fromgxb          : STD_LOGIC_VECTOR(c_reconf_fromgxb_bus_w*g_nof_gx-1 DOWNTO 0);

  SIGNAL rx_freq_locked            : STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0) := (OTHERS=>'0');
  SIGNAL tx_pll_locked             : STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0) := (OTHERS=>'0');
  SIGNAL rec_busy                  : STD_LOGIC;
  
  SIGNAL trc_tx_pll_powerdown      : STD_LOGIC;
  SIGNAL trc_tx_digital_rst        : STD_LOGIC;

  SIGNAL trc_rx_analog_rst         : STD_LOGIC;
  SIGNAL trc_rx_digital_rst        : STD_LOGIC;

  SIGNAL i_tx_clk                  : STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
  SIGNAL i_rx_clk                  : STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);

  SIGNAL i_tx_rst                  : STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
  SIGNAL i_rx_rst                  : STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
 
  SIGNAL tr_rst                    : STD_LOGIC;

  SIGNAL trc_tx_rdy                : STD_LOGIC;
  SIGNAL trc_rx_rdy                : STD_LOGIC;

  SIGNAL not_trc_tx_rdy            : STD_LOGIC;
  SIGNAL not_trc_rx_rdy            : STD_LOGIC;

  SIGNAL tx_align_en_out           : STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0) := (OTHERS=>'0');

  SIGNAL tx_datain                 : t_data_arr(g_nof_gx-1 DOWNTO 0);
  SIGNAL rx_dataout                : t_data_arr(g_nof_gx-1 DOWNTO 0);

  SIGNAL txc_tx_ctrlenable         : t_ctrl_arr(g_nof_gx-1 DOWNTO 0);
  SIGNAL rxc_rx_ctrldetect         : t_ctrl_arr(g_nof_gx-1 DOWNTO 0); 

  SIGNAL rxc_rx_align_en           : STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0); --edge sensitive
  SIGNAL rxc_rx_aligned            : STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0) := (OTHERS=>'1');
  
BEGIN
 
  ------------------------------------------------------------------------------
  -- ALTGX Megafunction
  ------------------------------------------------------------------------------

  gen_altgx: FOR i IN 0 to g_nof_gx-1 GENERATE  --Generate g_nof_gx one-channel ALTGX instances 
  
    gen_tr : IF g_tx = TRUE AND g_rx = TRUE GENERATE  --Generate duplex transceivers
    
      gen_32b : IF g_data_w = 32 GENERATE
        u_gx: ip_stratixiv_hssi_gx_32b_generic
        GENERIC MAP (
          g_mbps                   => g_mbps,
          starting_channel_number  => i*4  --Starting channel numbers must be 0,4,8,12 etc for individual ALTGX instances
        )                                  --Reconfig megawizard: regenerate with 'number of channels' = 4*nof_gx 
        PORT MAP (   
          -- Clocks, in
          pll_inclk           => tr_clk,  
          cal_blk_clk         => cal_rec_clk,      
          reconfig_clk        => cal_rec_clk,
        
          -- Clocks, generated (out)
          sl(tx_clkout)       => i_tx_clk(i),   
          sl(rx_clkout)       => i_rx_clk(i), 
        
          -- Parallel data I/O
          rx_dataout          => rx_sosi_arr(i).data(g_data_w-1 DOWNTO 0), 
          tx_datain           => tx_datain(i), 
        
          -- Serial data I/O 
          rx_datain           => slv(rx_datain(i)),  
          sl(tx_dataout)      => tx_dataout(i),
        
          -- Initialization I/O
          sl(rx_freqlocked)   => rx_freq_locked(i),
          sl(pll_locked)      => tx_pll_locked(i),
          pll_powerdown       => slv(trc_tx_pll_powerdown),  
          rx_digitalreset     => slv(trc_rx_digital_rst),                  
          rx_analogreset      => slv(trc_rx_analog_rst),
          tx_digitalreset     => slv(trc_tx_digital_rst),
        
          -- 8b/10b related
          tx_ctrlenable       => txc_tx_ctrlenable(i), --8b10b
          rx_ctrldetect       => rxc_rx_ctrldetect(i), --8b10b
          rx_enapatternalign  => slv(rxc_rx_align_en(i)),
        
          -- Reconfig block connections
          reconfig_togxb      => reconfig_togxb,
          reconfig_fromgxb    => reconfig_fromgxb(c_reconf_fromgxb_bus_w*(i+1)-1 DOWNTO c_reconf_fromgxb_bus_w*(i+1)-c_reconf_fromgxb_bus_w),
        
          rx_seriallpbken     => slv('0'),
          gxb_powerdown       => (OTHERS=>'0')
        );  
      END GENERATE;

      gen_16b : IF g_data_w = 16 GENERATE 
        u_gx: ip_stratixiv_hssi_gx_16b
        GENERIC MAP (
          starting_channel_number  => i*4  --Starting channel numbers must be 0,4,8,12 etc for individual ALTGX instances
        )                                  --Reconfig megawizard: regenerate with 'number of channels' = 4*nof_gx 
        PORT MAP (   
          -- Clocks, in
          pll_inclk           => tr_clk,  
          cal_blk_clk         => cal_rec_clk,      
          reconfig_clk        => cal_rec_clk,
     
          -- Clocks, generated (out)
          sl(tx_clkout)       => i_tx_clk(i),   
          sl(rx_clkout)       => i_rx_clk(i), 
  
          -- Parallel data I/O
          rx_dataout          => rx_dataout(i), 
          tx_datain           => tx_datain(i), 
  
          -- Serial data I/O 
          rx_datain           => slv(rx_datain(i)),  
          sl(tx_dataout)      => tx_dataout(i),
  
          -- Initialization I/O
          sl(rx_freqlocked)   => rx_freq_locked(i),
          sl(pll_locked)      => tx_pll_locked(i),
          pll_powerdown       => slv(trc_tx_pll_powerdown),  
          rx_digitalreset     => slv(trc_rx_digital_rst),                  
          rx_analogreset      => slv(trc_rx_analog_rst),
          tx_digitalreset     => slv(trc_tx_digital_rst),
  
          -- 8b/10b related
          tx_ctrlenable       => txc_tx_ctrlenable(i), --8b10b
          rx_ctrldetect       => rxc_rx_ctrldetect(i), --8b10b
   
          -- Reconfig block connections
          reconfig_togxb      => reconfig_togxb,
          reconfig_fromgxb    => reconfig_fromgxb(c_reconf_fromgxb_bus_w*(i+1)-1 DOWNTO c_reconf_fromgxb_bus_w*(i+1)-c_reconf_fromgxb_bus_w)
        );  
      END GENERATE;
    END GENERATE;    
    
    gen_rx : IF g_tx = FALSE AND g_rx = TRUE GENERATE  --Generate receivers only

      gen_32b : IF g_data_w = 32 GENERATE
        u_rx: ip_stratixiv_hssi_rx_32b_generic
        GENERIC MAP (
          g_mbps                   => g_mbps,
          starting_channel_number  => i*4  --Starting channel numbers must be 0,4,8,12 etc for individual ALTGX instances     
        )                                  --Reconfig megawizard: regenerate with 'number of channels' = 4*nof_gx
        PORT MAP (   
          cal_blk_clk         => cal_rec_clk,
          gxb_powerdown       => (OTHERS=>'0'),
          reconfig_clk        => cal_rec_clk,
          reconfig_togxb      => reconfig_togxb,
          rx_analogreset      => slv(trc_rx_analog_rst),
          rx_datain           => slv(rx_datain(i)),
          rx_digitalreset     => slv(trc_rx_digital_rst),
          reconfig_fromgxb    => reconfig_fromgxb(c_reconf_fromgxb_bus_w*(i+1)-1 DOWNTO c_reconf_fromgxb_bus_w*(i+1)-c_reconf_fromgxb_bus_w),
          sl(rx_clkout)       => i_rx_clk(i),
          rx_ctrldetect       => rxc_rx_ctrldetect(i),
          rx_dataout          => rx_sosi_arr(i).data(g_data_w-1 DOWNTO 0),
          sl(rx_freqlocked)   => rx_freq_locked(i),
          rx_enapatternalign  => slv(rxc_rx_align_en(i)),
          
          --ADDITTIONAL SIGNAL:
          rx_cruclk           => slv(tr_clk)
        );  
      END GENERATE;

      gen_16b : IF g_data_w = 16 GENERATE 
        u_rx: ip_stratixiv_hssi_rx_16b
        GENERIC MAP (
          starting_channel_number  => i*4  --Starting channel numbers must be 0,4,8,12 etc for individual ALTGX instances
        )                                  --Reconfig megawizard: regenerate with 'number of channels' = 4*nof_gx 
        PORT MAP (   
          -- Clocks, in
          cal_blk_clk         => cal_rec_clk,      
          reconfig_clk        => cal_rec_clk,
     
          -- Clocks, generated (out)
          sl(rx_clkout)       => i_rx_clk(i), 
  
          -- Parallel data I/O
          rx_dataout          => rx_dataout(i), 
  
          -- Serial data I/O 
          rx_datain           => slv(rx_datain(i)),  
  
          -- Initialization I/O
          sl(rx_freqlocked)   => rx_freq_locked(i),
          rx_digitalreset     => slv(trc_rx_digital_rst),                  
          rx_analogreset      => slv(trc_rx_analog_rst),
 
          -- 8b/10b related
          rx_ctrldetect       => rxc_rx_ctrldetect(i), --8b10b
   
          -- Reconfig block connections
          reconfig_togxb      => reconfig_togxb,
          reconfig_fromgxb    => reconfig_fromgxb(c_reconf_fromgxb_bus_w*(i+1)-1 DOWNTO c_reconf_fromgxb_bus_w*(i+1)-c_reconf_fromgxb_bus_w),

          --ADDITTIONAL SIGNAL:
          rx_cruclk           => slv(tr_clk)
        );  
      END GENERATE;
    END GENERATE;
    
    gen_tx : IF g_tx = TRUE AND g_rx = FALSE GENERATE
      gen_32b : IF g_data_w = 32 GENERATE
        u_tx: ip_stratixiv_hssi_tx_32b_generic
        GENERIC MAP (
           g_mbps             => g_mbps
        )  
        PORT MAP (   
          cal_blk_clk         => cal_rec_clk,
          gxb_powerdown       => (OTHERS=>'0'),
          pll_inclk           => tr_clk,
          pll_powerdown       => slv(trc_tx_pll_powerdown),
          tx_ctrlenable       => txc_tx_ctrlenable(i),
          tx_datain           => tx_datain(i),
          tx_digitalreset     => slv(trc_tx_digital_rst),
          sl(pll_locked)      => tx_pll_locked(i),
          sl(tx_clkout)       => i_tx_clk(i),     
          sl(tx_dataout)      => tx_dataout(i)
        );  
      END GENERATE; 

      gen_16b : IF g_data_w = 16 GENERATE 
        u_tx: ip_stratixiv_hssi_tx_16b
        PORT MAP (   
          -- Clocks, in
          pll_inclk           => tr_clk,  
          cal_blk_clk         => cal_rec_clk,      
     
          -- Clocks, generated (out)
          sl(tx_clkout)       => i_tx_clk(i),   
  
          -- Parallel data I/O
          tx_datain           => tx_datain(i), 
  
          -- Serial data I/O 
          sl(tx_dataout)      => tx_dataout(i),
  
          -- Initialization I/O
          sl(pll_locked)      => tx_pll_locked(i),
          pll_powerdown       => slv(trc_tx_pll_powerdown),                
          tx_digitalreset     => slv(trc_tx_digital_rst),
  
          -- 8b/10b related
          tx_ctrlenable       => txc_tx_ctrlenable(i) --8b10b   
        );  
      END GENERATE;
    END GENERATE;
  END GENERATE;

  
  ------------------------------------------------------------------------------
  -- Reset and alignment
  ------------------------------------------------------------------------------
  
  u_areset_tr_rst : ENTITY common_lib.common_areset
  GENERIC MAP(
    g_rst_level => '1'
  )
  PORT MAP(
    clk    => tr_clk,
    in_rst => '0',
    out_rst => tr_rst
  );  

  gen_tx : IF g_tx = TRUE GENERATE
 
    tx_clk <= i_tx_clk;
    tx_rst <= i_tx_rst;

    not_trc_tx_rdy <= NOT(trc_tx_rdy);

    u_tech_transceiver_tx_rst : ENTITY work.tech_transceiver_tx_rst
    GENERIC MAP(
      g_nof_gx => g_nof_gx
    )
    PORT MAP (
      tr_clk               => tr_clk,
      tr_rst               => tr_rst,
      trc_tx_pll_powerdown => trc_tx_pll_powerdown,
      tx_pll_locked        => tx_pll_locked,
      trc_tx_digital_rst   => trc_tx_digital_rst,
      trc_tx_rdy           => trc_tx_rdy
    );

    gen_altgx: FOR i IN 0 to g_nof_gx-1 GENERATE

      u_areset_tx_rst : ENTITY common_lib.common_areset
      GENERIC MAP(
        g_rst_level => '1'
      )
      PORT MAP(
        clk     => i_tx_clk(i),
        in_rst  => not_trc_tx_rdy,
        out_rst => i_tx_rst(i)
      );
      
      -- ALTGX in 16-bit mode supports and uses internal auto alignment state machine.
      gen_tx_align: IF g_data_w=32 GENERATE
        u_tech_transceiver_tx_align : ENTITY work.tech_transceiver_tx_align
        GENERIC MAP(
          g_sim    => g_sim,
          g_data_w => g_data_w,
          g_mbps   => g_mbps
        )
        PORT MAP (
          tx_clk             => i_tx_clk(i), 
          tx_rst             => i_tx_rst(i),
          tx_align_en_in     => tx_align_en(i),
          tx_align_en_out    => tx_align_en_out(i),
          tx_state           => tx_state(2*i+1 DOWNTO 2*i)
        );
      END GENERATE;

      -- TX is ready when it's not sending the alignment pattern
      tx_siso_arr(i).ready     <= NOT tx_align_en_out(i);
      tx_siso_arr(i).xon       <= NOT tx_align_en_out(i);
      tx_datain(i)             <= c_align_pattern WHEN tx_align_en_out(i) = '1' ELSE tx_sosi_arr(i).data(g_data_w-1 DOWNTO 0) WHEN  tx_sosi_arr(i).valid = '1' ELSE c_inval_pattern;
      txc_tx_ctrlenable(i)     <= (OTHERS=>'1') WHEN tx_align_en_out(i) = '1' ELSE (OTHERS=>'0') WHEN tx_sosi_arr(i).valid = '1' ELSE (OTHERS=>'1');

    END GENERATE;
  END GENERATE;

  gen_rx : IF g_rx = TRUE GENERATE

    rx_clk <= i_rx_clk;  
    rx_rst <= i_rx_rst;

    u_tech_transceiver_rx_rst : ENTITY work.tech_transceiver_rx_rst
    GENERIC MAP(
      g_nof_gx => g_nof_gx
    )
    PORT MAP (
      tr_clk             => tr_clk,
      tr_rst             => tr_rst,
      rec_busy           => rec_busy,
      rx_freq_locked     => rx_freq_locked,
      trc_rx_analog_rst  => trc_rx_analog_rst,             
      trc_rx_digital_rst => trc_rx_digital_rst,  
      trc_rx_rdy         => trc_rx_rdy
    );

    not_trc_rx_rdy <= NOT(trc_rx_rdy);

    i_rx_rst_align: FOR i IN 0 to g_nof_gx-1 GENERATE

      u_areset_i_rx_rst : ENTITY common_lib.common_areset
      GENERIC MAP(
        g_rst_level => '1'
      )
      PORT MAP(
        clk     => i_rx_clk(i),
        in_rst  => not_trc_rx_rdy,
        out_rst => i_rx_rst(i)
      );

      -- ALTGX in 16-bit mode supports and uses internal auto alignment state machine.
      gen_rx_align: IF g_data_w=32 GENERATE
        u_tech_transceiver_rx_align : ENTITY work.tech_transceiver_rx_align
        GENERIC MAP(
          g_sim    => g_sim,
          g_data_w => g_data_w,
          g_mbps   => g_mbps
        )
        PORT MAP (
          rx_clk             => i_rx_clk(i), 
          rx_rst             => i_rx_rst(i),
          rx_align_en_in     => rx_align_en(i),
          rx_align_en_out    => rxc_rx_align_en(i), 
          rxc_rx_aligned     => rxc_rx_aligned(i),
          rx_state           => rx_state(2*i+1 DOWNTO 2*i)
        );

        -- The transmitters will represent gaps as the control-encoded c_inval pattern
        rx_sosi_arr(i).valid <= NOT(orv(rxc_rx_ctrldetect(i))) WHEN rxc_rx_aligned(i) = '1' ELSE '0';
      END GENERATE;

      gen_rx_order: IF g_data_w=16 GENERATE
        u_tech_transceiver_rx_align : ENTITY work.tech_transceiver_rx_order
        GENERIC MAP(
          g_data_w    => g_data_w
        )
        PORT MAP (
          rx_clk       => i_rx_clk(i), 
          rx_rst       => i_rx_rst(i),

          rx_data_in   => rx_dataout(i),
          rx_ctrl_in   => rxc_rx_ctrldetect(i),                     
                              
          rx_data_out  => rx_sosi_arr(i).data(g_data_w-1 DOWNTO 0),
          rx_valid_out => rx_sosi_arr(i).valid
        );
      END GENERATE;     
    END GENERATE;
  END GENERATE;


  ------------------------------------------------------------------------------
  -- ALTGX_RECONFIG Megafunction
  ------------------------------------------------------------------------------
    
  u_gxb_reconfig : ip_stratixiv_gxb_reconfig_v91  -- Create one gxb_reconfig module for all ALTGX instances
  GENERIC MAP (
    g_nof_gx        => g_nof_gx,
    g_fromgxb_bus_w => c_reconf_fromgxb_bus_w,  -- = 17
    g_togxb_bus_w   => c_reconf_togxb_bus_w     -- = 4
  )
  PORT MAP (
    reconfig_clk      => cal_rec_clk,
    reconfig_fromgxb  => reconfig_fromgxb,
    busy              => rec_busy,
    reconfig_togxb    => reconfig_togxb
  );
  
END str;

